"""Source datasets from Kaggle API.

See https://github.com/Kaggle/kaggle-api for more details.
"""

from collections import OrderedDict

from . import kaggle_regression, kaggle_classification


def classification_envs(n=5, test_size=None, random_state=None, verbose=False):
    # TODO: limit number of envs by n
    competitions = [
        kaggle_classification.homesite_quote_conversion(),
        kaggle_classification.santander_customer_satisfaction(),
        kaggle_classification.bnp_paribas_cardif_claims_management(),
        kaggle_classification.poker_rule_induction(),
        kaggle_classification.costa_rican_household_poverty_prediction(),
    ]
    return OrderedDict([
        (d.dataset_name, d.data_env(
            test_size=test_size, random_state=random_state))
        for d in competitions])


def regression_envs(n=5, test_size=None, random_state=None, verbose=False):
    # TODO: limit number of envs by n
    competitions = [
        kaggle_regression.restaurant_revenue_prediction(),
        kaggle_regression.nyc_taxi_trip_duration(),
        kaggle_regression.mercedes_benz_greener_manufacturing(),
        kaggle_regression.allstate_claims_severity(),
        kaggle_regression.house_prices_advanced_regression_techniques(),
    ]
    return OrderedDict([
        (d.dataset_name, d.data_env(
            test_size=test_size, random_state=random_state))
        for d in competitions])



"""Feature map for the kaggle competition 'santander customer satisfaction."""

from collections import OrderedDict

from ...data_types import FeatureType


def get_feature_map():
    return OrderedDict([
        ("delta_imp_amort_var18_1y3", FeatureType.CONTINUOUS),
        ("delta_imp_amort_var34_1y3", FeatureType.CONTINUOUS),
        ("delta_imp_aport_var13_1y3", FeatureType.CONTINUOUS),
        ("delta_imp_aport_var17_1y3", FeatureType.CONTINUOUS),
        ("delta_imp_aport_var33_1y3", FeatureType.CONTINUOUS),
        ("delta_imp_compra_var44_1y3", FeatureType.CONTINUOUS),
        ("delta_imp_reemb_var13_1y3", FeatureType.CONTINUOUS),
        ("delta_imp_reemb_var17_1y3", FeatureType.CONTINUOUS),
        ("delta_imp_reemb_var33_1y3", FeatureType.CONTINUOUS),
        ("delta_imp_trasp_var17_in_1y3", FeatureType.CONTINUOUS),
        ("delta_imp_trasp_var17_out_1y3", FeatureType.CONTINUOUS),
        ("delta_imp_trasp_var33_in_1y3", FeatureType.CONTINUOUS),
        ("delta_imp_trasp_var33_out_1y3", FeatureType.CONTINUOUS),
        ("delta_imp_venta_var44_1y3", FeatureType.CONTINUOUS),
        ("delta_num_aport_var13_1y3", FeatureType.CONTINUOUS),
        ("delta_num_aport_var17_1y3", FeatureType.CONTINUOUS),
        ("delta_num_aport_var33_1y3", FeatureType.CONTINUOUS),
        ("delta_num_compra_var44_1y3", FeatureType.CONTINUOUS),
        ("delta_num_reemb_var13_1y3", FeatureType.CONTINUOUS),
        ("delta_num_reemb_var17_1y3", FeatureType.CONTINUOUS),
        ("delta_num_reemb_var33_1y3", FeatureType.CONTINUOUS),
        ("delta_num_trasp_var17_in_1y3", FeatureType.CONTINUOUS),
        ("delta_num_trasp_var17_out_1y3", FeatureType.CONTINUOUS),
        ("delta_num_trasp_var33_in_1y3", FeatureType.CONTINUOUS),
        ("delta_num_trasp_var33_out_1y3", FeatureType.CONTINUOUS),
        ("delta_num_venta_var44_1y3", FeatureType.CONTINUOUS),
        ("imp_amort_var18_hace3", FeatureType.CONTINUOUS),
        ("imp_amort_var18_ult1", FeatureType.CONTINUOUS),
        ("imp_amort_var34_hace3", FeatureType.CONTINUOUS),
        ("imp_amort_var34_ult1", FeatureType.CONTINUOUS),
        ("imp_aport_var13_hace3", FeatureType.CONTINUOUS),
        ("imp_aport_var13_ult1", FeatureType.CONTINUOUS),
        ("imp_aport_var17_hace3", FeatureType.CONTINUOUS),
        ("imp_aport_var17_ult1", FeatureType.CONTINUOUS),
        ("imp_aport_var33_hace3", FeatureType.CONTINUOUS),
        ("imp_aport_var33_ult1", FeatureType.CONTINUOUS),
        ("imp_compra_var44_hace3", FeatureType.CONTINUOUS),
        ("imp_compra_var44_ult1", FeatureType.CONTINUOUS),
        ("imp_ent_var16_ult1", FeatureType.CONTINUOUS),
        ("imp_op_var39_comer_ult1", FeatureType.CONTINUOUS),
        ("imp_op_var39_comer_ult3", FeatureType.CONTINUOUS),
        ("imp_op_var39_efect_ult1", FeatureType.CONTINUOUS),
        ("imp_op_var39_efect_ult3", FeatureType.CONTINUOUS),
        ("imp_op_var39_ult1", FeatureType.CONTINUOUS),
        ("imp_op_var40_comer_ult1", FeatureType.CONTINUOUS),
        ("imp_op_var40_comer_ult3", FeatureType.CONTINUOUS),
        ("imp_op_var40_efect_ult1", FeatureType.CONTINUOUS),
        ("imp_op_var40_efect_ult3", FeatureType.CONTINUOUS),
        ("imp_op_var40_ult1", FeatureType.CONTINUOUS),
        ("imp_op_var41_comer_ult1", FeatureType.CONTINUOUS),
        ("imp_op_var41_comer_ult3", FeatureType.CONTINUOUS),
        ("imp_op_var41_efect_ult1", FeatureType.CONTINUOUS),
        ("imp_op_var41_efect_ult3", FeatureType.CONTINUOUS),
        ("imp_op_var41_ult1", FeatureType.CONTINUOUS),
        ("imp_reemb_var13_hace3", FeatureType.CONTINUOUS),
        ("imp_reemb_var13_ult1", FeatureType.CONTINUOUS),
        ("imp_reemb_var17_hace3", FeatureType.CONTINUOUS),
        ("imp_reemb_var17_ult1", FeatureType.CONTINUOUS),
        ("imp_reemb_var33_hace3", FeatureType.CONTINUOUS),
        ("imp_reemb_var33_ult1", FeatureType.CONTINUOUS),
        ("imp_sal_var16_ult1", FeatureType.CONTINUOUS),
        ("imp_trans_var37_ult1", FeatureType.CONTINUOUS),
        ("imp_trasp_var17_in_hace3", FeatureType.CONTINUOUS),
        ("imp_trasp_var17_in_ult1", FeatureType.CONTINUOUS),
        ("imp_trasp_var17_out_hace3", FeatureType.CONTINUOUS),
        ("imp_trasp_var17_out_ult1", FeatureType.CONTINUOUS),
        ("imp_trasp_var33_in_hace3", FeatureType.CONTINUOUS),
        ("imp_trasp_var33_in_ult1", FeatureType.CONTINUOUS),
        ("imp_trasp_var33_out_hace3", FeatureType.CONTINUOUS),
        ("imp_trasp_var33_out_ult1", FeatureType.CONTINUOUS),
        ("imp_var43_emit_ult1", FeatureType.CONTINUOUS),
        ("imp_var7_emit_ult1", FeatureType.CONTINUOUS),
        ("imp_var7_recib_ult1", FeatureType.CONTINUOUS),
        ("imp_venta_var44_hace3", FeatureType.CONTINUOUS),
        ("imp_venta_var44_ult1", FeatureType.CONTINUOUS),
        ("ind_var1", FeatureType.CONTINUOUS),
        ("ind_var10_ult1", FeatureType.CONTINUOUS),
        ("ind_var10cte_ult1", FeatureType.CONTINUOUS),
        ("ind_var12", FeatureType.CONTINUOUS),
        ("ind_var12_0", FeatureType.CONTINUOUS),
        ("ind_var13", FeatureType.CONTINUOUS),
        ("ind_var13_0", FeatureType.CONTINUOUS),
        ("ind_var13_corto", FeatureType.CONTINUOUS),
        ("ind_var13_corto_0", FeatureType.CONTINUOUS),
        ("ind_var13_largo", FeatureType.CONTINUOUS),
        ("ind_var13_largo_0", FeatureType.CONTINUOUS),
        ("ind_var13_medio", FeatureType.CONTINUOUS),
        ("ind_var13_medio_0", FeatureType.CONTINUOUS),
        ("ind_var14", FeatureType.CONTINUOUS),
        ("ind_var14_0", FeatureType.CONTINUOUS),
        ("ind_var17", FeatureType.CONTINUOUS),
        ("ind_var17_0", FeatureType.CONTINUOUS),
        ("ind_var18", FeatureType.CONTINUOUS),
        ("ind_var18_0", FeatureType.CONTINUOUS),
        ("ind_var19", FeatureType.CONTINUOUS),
        ("ind_var1_0", FeatureType.CONTINUOUS),
        ("ind_var2", FeatureType.CONTINUOUS),
        ("ind_var20", FeatureType.CONTINUOUS),
        ("ind_var20_0", FeatureType.CONTINUOUS),
        ("ind_var24", FeatureType.CONTINUOUS),
        ("ind_var24_0", FeatureType.CONTINUOUS),
        ("ind_var25", FeatureType.CONTINUOUS),
        ("ind_var25_0", FeatureType.CONTINUOUS),
        ("ind_var25_cte", FeatureType.CONTINUOUS),
        ("ind_var26", FeatureType.CONTINUOUS),
        ("ind_var26_0", FeatureType.CONTINUOUS),
        ("ind_var26_cte", FeatureType.CONTINUOUS),
        ("ind_var27", FeatureType.CONTINUOUS),
        ("ind_var27_0", FeatureType.CONTINUOUS),
        ("ind_var28", FeatureType.CONTINUOUS),
        ("ind_var28_0", FeatureType.CONTINUOUS),
        ("ind_var29", FeatureType.CONTINUOUS),
        ("ind_var29_0", FeatureType.CONTINUOUS),
        ("ind_var2_0", FeatureType.CONTINUOUS),
        ("ind_var30", FeatureType.CONTINUOUS),
        ("ind_var30_0", FeatureType.CONTINUOUS),
        ("ind_var31", FeatureType.CONTINUOUS),
        ("ind_var31_0", FeatureType.CONTINUOUS),
        ("ind_var32", FeatureType.CONTINUOUS),
        ("ind_var32_0", FeatureType.CONTINUOUS),
        ("ind_var32_cte", FeatureType.CONTINUOUS),
        ("ind_var33", FeatureType.CONTINUOUS),
        ("ind_var33_0", FeatureType.CONTINUOUS),
        ("ind_var34", FeatureType.CONTINUOUS),
        ("ind_var34_0", FeatureType.CONTINUOUS),
        ("ind_var37", FeatureType.CONTINUOUS),
        ("ind_var37_0", FeatureType.CONTINUOUS),
        ("ind_var37_cte", FeatureType.CONTINUOUS),
        ("ind_var39", FeatureType.CONTINUOUS),
        ("ind_var39_0", FeatureType.CONTINUOUS),
        ("ind_var40", FeatureType.CONTINUOUS),
        ("ind_var40_0", FeatureType.CONTINUOUS),
        ("ind_var41", FeatureType.CONTINUOUS),
        ("ind_var41_0", FeatureType.CONTINUOUS),
        ("ind_var43_emit_ult1", FeatureType.CONTINUOUS),
        ("ind_var43_recib_ult1", FeatureType.CONTINUOUS),
        ("ind_var44", FeatureType.CONTINUOUS),
        ("ind_var44_0", FeatureType.CONTINUOUS),
        ("ind_var46", FeatureType.CONTINUOUS),
        ("ind_var46_0", FeatureType.CONTINUOUS),
        ("ind_var5", FeatureType.CONTINUOUS),
        ("ind_var5_0", FeatureType.CONTINUOUS),
        ("ind_var6", FeatureType.CONTINUOUS),
        ("ind_var6_0", FeatureType.CONTINUOUS),
        ("ind_var7_emit_ult1", FeatureType.CONTINUOUS),
        ("ind_var7_recib_ult1", FeatureType.CONTINUOUS),
        ("ind_var8", FeatureType.CONTINUOUS),
        ("ind_var8_0", FeatureType.CONTINUOUS),
        ("ind_var9_cte_ult1", FeatureType.CONTINUOUS),
        ("ind_var9_ult1", FeatureType.CONTINUOUS),
        ("num_aport_var13_hace3", FeatureType.CONTINUOUS),
        ("num_aport_var13_ult1", FeatureType.CONTINUOUS),
        ("num_aport_var17_hace3", FeatureType.CONTINUOUS),
        ("num_aport_var17_ult1", FeatureType.CONTINUOUS),
        ("num_aport_var33_hace3", FeatureType.CONTINUOUS),
        ("num_aport_var33_ult1", FeatureType.CONTINUOUS),
        ("num_compra_var44_hace3", FeatureType.CONTINUOUS),
        ("num_compra_var44_ult1", FeatureType.CONTINUOUS),
        ("num_ent_var16_ult1", FeatureType.CONTINUOUS),
        ("num_med_var22_ult3", FeatureType.CONTINUOUS),
        ("num_med_var45_ult3", FeatureType.CONTINUOUS),
        ("num_meses_var12_ult3", FeatureType.CONTINUOUS),
        ("num_meses_var13_corto_ult3", FeatureType.CONTINUOUS),
        ("num_meses_var13_largo_ult3", FeatureType.CONTINUOUS),
        ("num_meses_var13_medio_ult3", FeatureType.CONTINUOUS),
        ("num_meses_var17_ult3", FeatureType.CONTINUOUS),
        ("num_meses_var29_ult3", FeatureType.CONTINUOUS),
        ("num_meses_var33_ult3", FeatureType.CONTINUOUS),
        ("num_meses_var39_vig_ult3", FeatureType.CONTINUOUS),
        ("num_meses_var44_ult3", FeatureType.CONTINUOUS),
        ("num_meses_var5_ult3", FeatureType.CONTINUOUS),
        ("num_meses_var8_ult3", FeatureType.CONTINUOUS),
        ("num_op_var39_comer_ult1", FeatureType.CONTINUOUS),
        ("num_op_var39_comer_ult3", FeatureType.CONTINUOUS),
        ("num_op_var39_efect_ult1", FeatureType.CONTINUOUS),
        ("num_op_var39_efect_ult3", FeatureType.CONTINUOUS),
        ("num_op_var39_hace2", FeatureType.CONTINUOUS),
        ("num_op_var39_hace3", FeatureType.CONTINUOUS),
        ("num_op_var39_ult1", FeatureType.CONTINUOUS),
        ("num_op_var39_ult3", FeatureType.CONTINUOUS),
        ("num_op_var40_comer_ult1", FeatureType.CONTINUOUS),
        ("num_op_var40_comer_ult3", FeatureType.CONTINUOUS),
        ("num_op_var40_efect_ult1", FeatureType.CONTINUOUS),
        ("num_op_var40_efect_ult3", FeatureType.CONTINUOUS),
        ("num_op_var40_hace2", FeatureType.CONTINUOUS),
        ("num_op_var40_hace3", FeatureType.CONTINUOUS),
        ("num_op_var40_ult1", FeatureType.CONTINUOUS),
        ("num_op_var40_ult3", FeatureType.CONTINUOUS),
        ("num_op_var41_comer_ult1", FeatureType.CONTINUOUS),
        ("num_op_var41_comer_ult3", FeatureType.CONTINUOUS),
        ("num_op_var41_efect_ult1", FeatureType.CONTINUOUS),
        ("num_op_var41_efect_ult3", FeatureType.CONTINUOUS),
        ("num_op_var41_hace2", FeatureType.CONTINUOUS),
        ("num_op_var41_hace3", FeatureType.CONTINUOUS),
        ("num_op_var41_ult1", FeatureType.CONTINUOUS),
        ("num_op_var41_ult3", FeatureType.CONTINUOUS),
        ("num_reemb_var13_hace3", FeatureType.CONTINUOUS),
        ("num_reemb_var13_ult1", FeatureType.CONTINUOUS),
        ("num_reemb_var17_hace3", FeatureType.CONTINUOUS),
        ("num_reemb_var17_ult1", FeatureType.CONTINUOUS),
        ("num_reemb_var33_hace3", FeatureType.CONTINUOUS),
        ("num_reemb_var33_ult1", FeatureType.CONTINUOUS),
        ("num_sal_var16_ult1", FeatureType.CONTINUOUS),
        ("num_trasp_var11_ult1", FeatureType.CONTINUOUS),
        ("num_trasp_var17_in_hace3", FeatureType.CONTINUOUS),
        ("num_trasp_var17_in_ult1", FeatureType.CONTINUOUS),
        ("num_trasp_var17_out_hace3", FeatureType.CONTINUOUS),
        ("num_trasp_var17_out_ult1", FeatureType.CONTINUOUS),
        ("num_trasp_var33_in_hace3", FeatureType.CONTINUOUS),
        ("num_trasp_var33_in_ult1", FeatureType.CONTINUOUS),
        ("num_trasp_var33_out_hace3", FeatureType.CONTINUOUS),
        ("num_trasp_var33_out_ult1", FeatureType.CONTINUOUS),
        ("num_var1", FeatureType.CONTINUOUS),
        ("num_var12", FeatureType.CONTINUOUS),
        ("num_var12_0", FeatureType.CONTINUOUS),
        ("num_var13", FeatureType.CONTINUOUS),
        ("num_var13_0", FeatureType.CONTINUOUS),
        ("num_var13_corto", FeatureType.CONTINUOUS),
        ("num_var13_corto_0", FeatureType.CONTINUOUS),
        ("num_var13_largo", FeatureType.CONTINUOUS),
        ("num_var13_largo_0", FeatureType.CONTINUOUS),
        ("num_var13_medio", FeatureType.CONTINUOUS),
        ("num_var13_medio_0", FeatureType.CONTINUOUS),
        ("num_var14", FeatureType.CONTINUOUS),
        ("num_var14_0", FeatureType.CONTINUOUS),
        ("num_var17", FeatureType.CONTINUOUS),
        ("num_var17_0", FeatureType.CONTINUOUS),
        ("num_var18", FeatureType.CONTINUOUS),
        ("num_var18_0", FeatureType.CONTINUOUS),
        ("num_var1_0", FeatureType.CONTINUOUS),
        ("num_var20", FeatureType.CONTINUOUS),
        ("num_var20_0", FeatureType.CONTINUOUS),
        ("num_var22_hace2", FeatureType.CONTINUOUS),
        ("num_var22_hace3", FeatureType.CONTINUOUS),
        ("num_var22_ult1", FeatureType.CONTINUOUS),
        ("num_var22_ult3", FeatureType.CONTINUOUS),
        ("num_var24", FeatureType.CONTINUOUS),
        ("num_var24_0", FeatureType.CONTINUOUS),
        ("num_var25", FeatureType.CONTINUOUS),
        ("num_var25_0", FeatureType.CONTINUOUS),
        ("num_var26", FeatureType.CONTINUOUS),
        ("num_var26_0", FeatureType.CONTINUOUS),
        ("num_var27", FeatureType.CONTINUOUS),
        ("num_var27_0", FeatureType.CONTINUOUS),
        ("num_var28", FeatureType.CONTINUOUS),
        ("num_var28_0", FeatureType.CONTINUOUS),
        ("num_var29", FeatureType.CONTINUOUS),
        ("num_var29_0", FeatureType.CONTINUOUS),
        ("num_var2_0_ult1", FeatureType.CONTINUOUS),
        ("num_var2_ult1", FeatureType.CONTINUOUS),
        ("num_var30", FeatureType.CONTINUOUS),
        ("num_var30_0", FeatureType.CONTINUOUS),
        ("num_var31", FeatureType.CONTINUOUS),
        ("num_var31_0", FeatureType.CONTINUOUS),
        ("num_var32", FeatureType.CONTINUOUS),
        ("num_var32_0", FeatureType.CONTINUOUS),
        ("num_var33", FeatureType.CONTINUOUS),
        ("num_var33_0", FeatureType.CONTINUOUS),
        ("num_var34", FeatureType.CONTINUOUS),
        ("num_var34_0", FeatureType.CONTINUOUS),
        ("num_var35", FeatureType.CONTINUOUS),
        ("num_var37", FeatureType.CONTINUOUS),
        ("num_var37_0", FeatureType.CONTINUOUS),
        ("num_var37_med_ult2", FeatureType.CONTINUOUS),
        ("num_var39", FeatureType.CONTINUOUS),
        ("num_var39_0", FeatureType.CONTINUOUS),
        ("num_var4", FeatureType.CONTINUOUS),
        ("num_var40", FeatureType.CONTINUOUS),
        ("num_var40_0", FeatureType.CONTINUOUS),
        ("num_var41", FeatureType.CONTINUOUS),
        ("num_var41_0", FeatureType.CONTINUOUS),
        ("num_var42", FeatureType.CONTINUOUS),
        ("num_var42_0", FeatureType.CONTINUOUS),
        ("num_var43_emit_ult1", FeatureType.CONTINUOUS),
        ("num_var43_recib_ult1", FeatureType.CONTINUOUS),
        ("num_var44", FeatureType.CONTINUOUS),
        ("num_var44_0", FeatureType.CONTINUOUS),
        ("num_var45_hace2", FeatureType.CONTINUOUS),
        ("num_var45_hace3", FeatureType.CONTINUOUS),
        ("num_var45_ult1", FeatureType.CONTINUOUS),
        ("num_var45_ult3", FeatureType.CONTINUOUS),
        ("num_var46", FeatureType.CONTINUOUS),
        ("num_var46_0", FeatureType.CONTINUOUS),
        ("num_var5", FeatureType.CONTINUOUS),
        ("num_var5_0", FeatureType.CONTINUOUS),
        ("num_var6", FeatureType.CONTINUOUS),
        ("num_var6_0", FeatureType.CONTINUOUS),
        ("num_var7_emit_ult1", FeatureType.CONTINUOUS),
        ("num_var7_recib_ult1", FeatureType.CONTINUOUS),
        ("num_var8", FeatureType.CONTINUOUS),
        ("num_var8_0", FeatureType.CONTINUOUS),
        ("num_venta_var44_hace3", FeatureType.CONTINUOUS),
        ("num_venta_var44_ult1", FeatureType.CONTINUOUS),
        ("saldo_medio_var12_hace2", FeatureType.CONTINUOUS),
        ("saldo_medio_var12_hace3", FeatureType.CONTINUOUS),
        ("saldo_medio_var12_ult1", FeatureType.CONTINUOUS),
        ("saldo_medio_var12_ult3", FeatureType.CONTINUOUS),
        ("saldo_medio_var13_corto_hace2", FeatureType.CONTINUOUS),
        ("saldo_medio_var13_corto_hace3", FeatureType.CONTINUOUS),
        ("saldo_medio_var13_corto_ult1", FeatureType.CONTINUOUS),
        ("saldo_medio_var13_corto_ult3", FeatureType.CONTINUOUS),
        ("saldo_medio_var13_largo_hace2", FeatureType.CONTINUOUS),
        ("saldo_medio_var13_largo_hace3", FeatureType.CONTINUOUS),
        ("saldo_medio_var13_largo_ult1", FeatureType.CONTINUOUS),
        ("saldo_medio_var13_largo_ult3", FeatureType.CONTINUOUS),
        ("saldo_medio_var13_medio_hace2", FeatureType.CONTINUOUS),
        ("saldo_medio_var13_medio_hace3", FeatureType.CONTINUOUS),
        ("saldo_medio_var13_medio_ult1", FeatureType.CONTINUOUS),
        ("saldo_medio_var13_medio_ult3", FeatureType.CONTINUOUS),
        ("saldo_medio_var17_hace2", FeatureType.CONTINUOUS),
        ("saldo_medio_var17_hace3", FeatureType.CONTINUOUS),
        ("saldo_medio_var17_ult1", FeatureType.CONTINUOUS),
        ("saldo_medio_var17_ult3", FeatureType.CONTINUOUS),
        ("saldo_medio_var29_hace2", FeatureType.CONTINUOUS),
        ("saldo_medio_var29_hace3", FeatureType.CONTINUOUS),
        ("saldo_medio_var29_ult1", FeatureType.CONTINUOUS),
        ("saldo_medio_var29_ult3", FeatureType.CONTINUOUS),
        ("saldo_medio_var33_hace2", FeatureType.CONTINUOUS),
        ("saldo_medio_var33_hace3", FeatureType.CONTINUOUS),
        ("saldo_medio_var33_ult1", FeatureType.CONTINUOUS),
        ("saldo_medio_var33_ult3", FeatureType.CONTINUOUS),
        ("saldo_medio_var44_hace2", FeatureType.CONTINUOUS),
        ("saldo_medio_var44_hace3", FeatureType.CONTINUOUS),
        ("saldo_medio_var44_ult1", FeatureType.CONTINUOUS),
        ("saldo_medio_var44_ult3", FeatureType.CONTINUOUS),
        ("saldo_medio_var5_hace2", FeatureType.CONTINUOUS),
        ("saldo_medio_var5_hace3", FeatureType.CONTINUOUS),
        ("saldo_medio_var5_ult1", FeatureType.CONTINUOUS),
        ("saldo_medio_var5_ult3", FeatureType.CONTINUOUS),
        ("saldo_medio_var8_hace2", FeatureType.CONTINUOUS),
        ("saldo_medio_var8_hace3", FeatureType.CONTINUOUS),
        ("saldo_medio_var8_ult1", FeatureType.CONTINUOUS),
        ("saldo_medio_var8_ult3", FeatureType.CONTINUOUS),
        ("saldo_var1", FeatureType.CONTINUOUS),
        ("saldo_var12", FeatureType.CONTINUOUS),
        ("saldo_var13", FeatureType.CONTINUOUS),
        ("saldo_var13_corto", FeatureType.CONTINUOUS),
        ("saldo_var13_largo", FeatureType.CONTINUOUS),
        ("saldo_var13_medio", FeatureType.CONTINUOUS),
        ("saldo_var14", FeatureType.CONTINUOUS),
        ("saldo_var17", FeatureType.CONTINUOUS),
        ("saldo_var18", FeatureType.CONTINUOUS),
        ("saldo_var20", FeatureType.CONTINUOUS),
        ("saldo_var24", FeatureType.CONTINUOUS),
        ("saldo_var25", FeatureType.CONTINUOUS),
        ("saldo_var26", FeatureType.CONTINUOUS),
        ("saldo_var27", FeatureType.CONTINUOUS),
        ("saldo_var28", FeatureType.CONTINUOUS),
        ("saldo_var29", FeatureType.CONTINUOUS),
        ("saldo_var2_ult1", FeatureType.CONTINUOUS),
        ("saldo_var30", FeatureType.CONTINUOUS),
        ("saldo_var31", FeatureType.CONTINUOUS),
        ("saldo_var32", FeatureType.CONTINUOUS),
        ("saldo_var33", FeatureType.CONTINUOUS),
        ("saldo_var34", FeatureType.CONTINUOUS),
        ("saldo_var37", FeatureType.CONTINUOUS),
        ("saldo_var40", FeatureType.CONTINUOUS),
        ("saldo_var41", FeatureType.CONTINUOUS),
        ("saldo_var42", FeatureType.CONTINUOUS),
        ("saldo_var44", FeatureType.CONTINUOUS),
        ("saldo_var46", FeatureType.CONTINUOUS),
        ("saldo_var5", FeatureType.CONTINUOUS),
        ("saldo_var6", FeatureType.CONTINUOUS),
        ("saldo_var8", FeatureType.CONTINUOUS),
        ("var15", FeatureType.CONTINUOUS),
        ("var21", FeatureType.CONTINUOUS),
        ("var3", FeatureType.CONTINUOUS),
        ("var36", FeatureType.CONTINUOUS),
        ("var3", FeatureType.CONTINUOUS),
    ])



"""Feature Preprocessor components.

NOTE:
- In order to use estimators as feature selectors, you can use
  sklearn.feature_selection.SelectFromModel
- Many of the decisions to use/exclude certain hyperparameters, as well as
  ranges for numerical hyperparameters followed the work done by the
  auto-sklearn project:

  https://github.com/automl/auto-sklearn/tree/master/autosklearn/pipeline/components

TODO:
- SelectPercentile feature selector works for classification using the
  following parameters: ["chi2", "f_classif", "mutual_info"]
  - for regression, it's ["f_regression", "mutual_info"].
"""

from sklearn.cluster import FeatureAgglomeration
from sklearn.decomposition import PCA, FastICA, KernelPCA, TruncatedSVD
from sklearn.ensemble import (
    RandomTreesEmbedding, ExtraTreesClassifier, ExtraTreesRegressor)
from sklearn.feature_selection import (
    SelectFromModel, SelectPercentile, GenericUnivariateSelect,
    VarianceThreshold)
from sklearn.kernel_approximation import RBFSampler, Nystroem
from sklearn.preprocessing import PolynomialFeatures
from sklearn.svm import LinearSVC

from .algorithm_component import AlgorithmComponent, EXCLUDE_ALL
from .hyperparameter import (
    CategoricalHyperparameter, UniformIntHyperparameter,
    UniformFloatHyperparameter, TuplePairHyperparameter)
from . import constants


# constant used for rbf kernel
# https://github.com/automl/auto-sklearn/blob/master/autosklearn/pipeline/components/feature_preprocessing/kitchen_sinks.py
GAMMA_MIN = 3.0517578125e-05


def extra_trees_classification_preprocessor():
    # TODO: This component needs to know if the supervised task in the ML
    # framework is a classification or regression task
    pass


def extra_trees_regression_preprocessor():
    # TODO: This component needs to know if the supervised task in the ML
    # framework is a classification or regression task
    pass


def fast_ica():
    """Create FastICA component.

    Leave out the following hyperparameters:
    - fun_args
    - max_iter
    - tol
    - w_init
    - random_state
    """
    return AlgorithmComponent(
        name="FastICA",
        component_class=FastICA,
        component_type=constants.FEATURE_PREPROCESSOR,
        hyperparameters=[
            UniformIntHyperparameter(
                "n_components", 10, 2000, default=100),
            CategoricalHyperparameter(
                "algorithm", ["parallel", "deflation"], default="parallel"),
            CategoricalHyperparameter(
                "fun", ["logcosh", "exp", "cube"], default="logcosh"),
        ],
        constant_hyperparameters={
            "whiten": True
        })


def feature_agglomeration():
    """Create FeatureAgglomoration component."""
    return AlgorithmComponent(
        name="FeatureAgglomeration",
        component_class=FeatureAgglomeration,
        component_type=constants.FEATURE_PREPROCESSOR,
        hyperparameters=[
            UniformIntHyperparameter("n_clusters", 2, 400, default=2, n=20),
            CategoricalHyperparameter(
                "affinity", ["euclidean", "l1", "l2", "manhattan", "cosine"],
                default="euclidean"),
            CategoricalHyperparameter(
                "linkage", ["ward", "complete", "average"], default="ward"),
            CategoricalHyperparameter(
                "pooling_func", ["mean", "median", "max"], default="mean"),
        ])


def kernel_pca():
    """Create KernelPCA component.

    Leave out the following hyperaparameters:
    - kernel_params
    - alpha
    - fit_inverse_transform
    - eigen_solver
    - tol
    - max_iter
    - remove_zero_eig
    - random_state
    """
    return AlgorithmComponent(
        name="KernelPCA",
        component_class=KernelPCA,
        component_type=constants.FEATURE_PREPROCESSOR,
        hyperparameters=[
            UniformIntHyperparameter("n_components", 10, 2000, default=100),
            CategoricalHyperparameter(
                "kernel", ["poly", "rbf", "sigmoid", "cosine"], default="rbf"),
            UniformFloatHyperparameter(
                "gamma", GAMMA_MIN, 8.0, default=1.0, log=True),
            UniformIntHyperparameter("degree", 2, 5, 3),
            UniformIntHyperparameter("coef0", -1, 1, default=0),
        ],
        exclusion_conditions={
            "kernel": {
                "rbf": {
                    "degree": EXCLUDE_ALL,
                    "coef0": EXCLUDE_ALL,
                },
                "sigmoid": {
                    "degree": EXCLUDE_ALL,
                    "gamma": EXCLUDE_ALL,
                },
                "cosine": {
                    "degree": EXCLUDE_ALL,
                    "coef0": EXCLUDE_ALL,
                    "gamma": EXCLUDE_ALL,
                },
            }
        })


def rbf_sampler():
    """Create RBF Sampler 'Kitchen Sink' Component."""
    return AlgorithmComponent(
        name="RBFSample",
        component_class=RBFSampler,
        component_type=constants.FEATURE_PREPROCESSOR,
        hyperparameters=[
            UniformFloatHyperparameter("gamma", GAMMA_MIN, 8.0, default=1.0),
            UniformIntHyperparameter(
                "n_components", 5, 10000, default=100, log=True, n=10),
        ])


def liblinear_svc_preprocessor():
    # TODO: This component needs to know if the supervised task in the ML
    # framework is a classification or regression task
    pass


def nystroem_sampler():
    """Create Nystroem Component."""
    return AlgorithmComponent(
        name="Nystroem",
        component_class=Nystroem,
        component_type=constants.FEATURE_PREPROCESSOR,
        hyperparameters=[
            CategoricalHyperparameter(
                "kernel", ["poly", "rbf", "sigmoid", "cosine"], default="rbf"),
            UniformIntHyperparameter(
                "n_components", 5, 10000, default=100, log=True),
            UniformFloatHyperparameter(
                "gamma", GAMMA_MIN, 8.0, log=True, default=1.0),
            UniformIntHyperparameter("degree", 2, 5, 3),
            UniformFloatHyperparameter("coef0", -1, 1, default=0),
        ],
        exclusion_conditions={
            "kernel": {
                "rbf": {
                    "degree": EXCLUDE_ALL,
                    "coef0": EXCLUDE_ALL,
                },
                "cosine": {
                    "degree": EXCLUDE_ALL,
                    "coef0": EXCLUDE_ALL
                },
                "sigmoid": {
                    "degree": EXCLUDE_ALL
                },
            }
        })


def pca():
    """Create a PCA Component.

    Note here that we use floats for the n_components TuplePairHyperparameter,
    for more details see:

    http://scikit-learn.org/stable/modules/generated/sklearn.decomposition.PCA.html

    Leave out the following hyperparameters:
    - svd_solver: always use "full"
    - tol: only used when ``svd_solver == "arpack"``
    - iterated_power: only used when ``svd_solver == "randomized"``
    """
    return AlgorithmComponent(
        name="PCA",
        component_class=PCA,
        component_type=constants.FEATURE_PREPROCESSOR,
        hyperparameters=[
            UniformFloatHyperparameter(
                "n_components", 0.5, 0.999, default=0.999),
            CategoricalHyperparameter("whiten", [True, False], default=False)
        ])


def polynomial_features():
    """Create Polynomial Component."""
    return AlgorithmComponent(
        name="PolynomialFeatures",
        component_class=PolynomialFeatures,
        component_type=constants.FEATURE_PREPROCESSOR,
        hyperparameters=[
            UniformIntHyperparameter("degree", 2, 3, default=2, n=2),
            CategoricalHyperparameter(
                "interaction_only", [True, False], default=False),
            CategoricalHyperparameter(
                "include_bias", [True, False], default=True)
        ])


def random_trees_embedding():
    """Create RandomTreesEmbedding."""
    return AlgorithmComponent(
        name="RandomTreesEmbedding",
        component_class=RandomTreesEmbedding,
        component_type=constants.FEATURE_PREPROCESSOR,
        hyperparameters=[
            UniformIntHyperparameter("n_estimators", 10, 100, default=10),
            UniformIntHyperparameter("max_depth", 2, 10, default=5),
            UniformIntHyperparameter("min_samples_split", 2, 20, default=2),
            UniformIntHyperparameter("min_samples_leaf", 1, 20, default=1),
            UniformFloatHyperparameter(
                "min_weight_fraction_leaf", 0.0, 0.5, default=0.0),
            UniformIntHyperparameter(
                "max_leaf_nodes", 10, 1000, default=None, log=True),
            CategoricalHyperparameter(
                "sparse_output", [True, False], default=False)
        ],
        constant_hyperparameters={"sparse_output": False})


def select_percentile_classification():
    # TODO: This component needs to know if the supervised task in the ML
    # framework is a classification or regression task
    pass


def select_percentile_regression():
    # TODO: This component needs to know if the supervised task in the ML
    # framework is a classification or regression task
    pass


def truncated_svd():
    return AlgorithmComponent(
        name="TruncatedSVD",
        component_class=TruncatedSVD,
        component_type=constants.FEATURE_PREPROCESSOR,
        hyperparameters=[
            UniformIntHyperparameter("n_components", 10, 256, default=128)
        ])


def variance_threshold_filter():
    """Create a variance threshold filter component.

    Removes features that are below a threshold variance.
    """
    return AlgorithmComponent(
        name="VarianceThresholdFilter",
        component_class=VarianceThreshold,
        component_type=constants.FEATURE_PREPROCESSOR,
        hyperparameters=[
            UniformFloatHyperparameter("threshold", 0.0, 10.0, default=0.0)
        ])



"""Metadata for kaggle regression datasets."""

from collections import OrderedDict

from .kaggle_base import KaggleCompetition

from ..data_types import FeatureType, TargetType
from .. import scorers


def restaurant_revenue_prediction():
    """Create data interface to kaggle 'restaurant revenue prediction'.

    url: https://www.kaggle.com/c/restaurant-revenue-prediction
    """
    return KaggleCompetition(
        competition_id="restaurant-revenue-prediction",
        features=OrderedDict(
            [("Open Date", FeatureType.DATE),
             ("City", FeatureType.CATEGORICAL),
             ("City Group", FeatureType.CATEGORICAL),
             ("Type", FeatureType.CATEGORICAL)] +
            [("P%i" % (i + 1), FeatureType.CONTINUOUS) for i in range(37)]
        ),
        target={"revenue": TargetType.REGRESSION},
        training_data_fname="train.csv.zip",
        test_data_fname="test.csv.zip",
        custom_preprocessor=None,
        scorer=scorers.root_mean_squared_error())


def nyc_taxi_trip_duration():
    """Create data interface to kaggle 'nyc taxi trip duration'.

    url: https://www.kaggle.com/c/nyc-taxi-trip-duration
    """
    return KaggleCompetition(
        competition_id="nyc-taxi-trip-duration",
        features=OrderedDict([
            ("vendor_id", FeatureType.CATEGORICAL),
            ("pickup_datetime", FeatureType.DATE),
            ("passenger_count", FeatureType.CONTINUOUS),
            ("pickup_longitude", FeatureType.CONTINUOUS),
            ("pickup_latitude", FeatureType.CONTINUOUS),
            ("dropoff_longitude", FeatureType.CONTINUOUS),
            ("dropoff_latitude", FeatureType.CONTINUOUS),
            ("store_and_fwd_flag", FeatureType.CATEGORICAL)]),
        target={"trip_duration": TargetType.REGRESSION},
        training_data_fname="train.zip",
        test_data_fname="test.zip",
        custom_preprocessor=None,
        scorer=scorers.root_mean_squared_log_error())


def mercedes_benz_greener_manufacturing():
    """Create data interface to kaggle 'mercedes benz greener manufacturing'.

    url: https://www.kaggle.com/c/mercedes-benz-greener-manufacturing
    """
    # these column numbers are not in the datasset (it's unclear why)
    exclude_nums = [7, 25, 72, 121, 149, 188, 193, 303, 381]
    return KaggleCompetition(
        competition_id="mercedes-benz-greener-manufacturing",
        features=OrderedDict(
            # NOTE: column "X9" does not exist.
            [("X%i" % i, FeatureType.CATEGORICAL)
             for i in range(9) if i not in exclude_nums] +
            [("X%i" % i, FeatureType.CONTINUOUS)
             for i in range(10, 386) if i not in exclude_nums]),
        target={"y": TargetType.REGRESSION},
        training_data_fname="train.csv.zip",
        test_data_fname="test.csv.zip",
        custom_preprocessor=None,
        scorer=scorers.r2_score())


def allstate_claims_severity():
    """Create data interface to kaggle 'allstate claims severity'.

    url: https://www.kaggle.com/c/allstate-claims-severity
    """
    return KaggleCompetition(
        competition_id="allstate-claims-severity",
        features=OrderedDict(
            # NOTE: column "X9" does not exist.
            [("cat%i" % (i + 1), FeatureType.CATEGORICAL)
             for i in range(116)] +
            [("cont%i" % (i + 1), FeatureType.CONTINUOUS)
             for i in range(14)]),
        target={"loss": TargetType.REGRESSION},
        training_data_fname="train.csv.zip",
        test_data_fname="test.csv.zip",
        custom_preprocessor=None,
        scorer=scorers.mean_absolute_error())


def house_prices_advanced_regression_techniques():
    """Create data interface to 'house prices advanced regression techniques'.

    url: https://www.kaggle.com/c/house-prices-advanced-regression-techniques
    """
    categorical_features = [
        "MSSubClass",
        "MSZoning",
        "Street",
        "Alley",
        "LotShape",
        "LandContour",
        "Utilities",
        "LotConfig",
        "LandSlope",
        "Neighborhood",
        "Condition1",
        "Condition2",
        "BldgType",
        "HouseStyle",
        "OverallQual",
        "OverallCond",
        "RoofStyle",
        "RoofMatl",
        "Exterior1st",
        "Exterior2nd",
        "MasVnrType",
        "ExterQual",
        "ExterCond",
        "Foundation",
        "BsmtQual",
        "BsmtCond",
        "BsmtExposure",
        "BsmtFinType1",
        "BsmtFinType2",
        "Heating",
        "HeatingQC",
        "CentralAir",
        "Electrical",
        "KitchenQual",
        "Functional",
        "FireplaceQu",
        "GarageType",
        "GarageFinish",
        "GarageQual",
        "GarageCond",
        "PavedDrive",
        "PoolQC",
        "Fence",
        "MiscFeature",
        "SaleType",
        "SaleCondition",
    ]
    continuous_features = [
        "LotFrontage",
        "LotArea",
        "YearBuilt",
        "MasVnrArea",
        "BsmtFinSF1",
        "BsmtFinSF2",
        "BsmtUnfSF",
        "TotalBsmtSF",
        "1stFlrSF",
        "2ndFlrSF",
        "LowQualFinSF",
        "GrLivArea",
        "BsmtFullBath",
        "BsmtHalfBath",
        "FullBath",
        "HalfBath",
        "TotRmsAbvGrd",
        "Fireplaces",
        "GarageYrBlt",
        "GarageCars",
        "GarageArea",
        "WoodDeckSF",
        "EnclosedPorch",
        "3SsnPorch",
        "ScreenPorch",
        "PoolArea",
        "MiscVal",
        "YrSold",
    ]
    return KaggleCompetition(
        competition_id="house-prices-advanced-regression-techniques",
        features=OrderedDict(
            [(c, FeatureType.CATEGORICAL) for c in categorical_features] +
            [(c, FeatureType.CONTINUOUS) for c in continuous_features]),
        target={"SalePrice": TargetType.REGRESSION},
        training_data_fname="train.csv.gz",
        test_data_fname="test.csv.gz",
        custom_preprocessor=None,
        scorer=scorers.root_mean_squared_log_error())



"""Feature map for the kaggle competition 'homesite quote conversion."""

from collections import OrderedDict

from ...data_types import FeatureType


def get_feature_map():
    return OrderedDict([
        ("Original_Quote_Date", FeatureType.DATE),
        ("Field6", FeatureType.CATEGORICAL),
        ("Field7", FeatureType.CONTINUOUS),
        ("Field8", FeatureType.CONTINUOUS),
        ("Field9", FeatureType.CONTINUOUS),
        ("Field10", FeatureType.CONTINUOUS),
        ("Field11", FeatureType.CONTINUOUS),
        ("Field12", FeatureType.CATEGORICAL),
        ("CoverageField1A", FeatureType.CONTINUOUS),
        ("CoverageField1B", FeatureType.CONTINUOUS),
        ("CoverageField2A", FeatureType.CONTINUOUS),
        ("CoverageField2B", FeatureType.CONTINUOUS),
        ("CoverageField3A", FeatureType.CONTINUOUS),
        ("CoverageField3B", FeatureType.CONTINUOUS),
        ("CoverageField4A", FeatureType.CONTINUOUS),
        ("CoverageField4B", FeatureType.CONTINUOUS),
        ("CoverageField5A", FeatureType.CONTINUOUS),
        ("CoverageField5B", FeatureType.CONTINUOUS),
        ("CoverageField6A", FeatureType.CONTINUOUS),
        ("CoverageField6B", FeatureType.CONTINUOUS),
        ("CoverageField8", FeatureType.CATEGORICAL),
        ("CoverageField9", FeatureType.CATEGORICAL),
        ("CoverageField11A", FeatureType.CONTINUOUS),
        ("CoverageField11B", FeatureType.CONTINUOUS),
        ("SalesField1A", FeatureType.CONTINUOUS),
        ("SalesField1B", FeatureType.CONTINUOUS),
        ("SalesField2A", FeatureType.CONTINUOUS),
        ("SalesField2B", FeatureType.CONTINUOUS),
        ("SalesField3", FeatureType.CONTINUOUS),
        ("SalesField4", FeatureType.CONTINUOUS),
        ("SalesField5", FeatureType.CONTINUOUS),
        ("SalesField6", FeatureType.CONTINUOUS),
        ("SalesField7", FeatureType.CATEGORICAL),
        ("SalesField8", FeatureType.CONTINUOUS),
        ("SalesField9", FeatureType.CONTINUOUS),
        ("SalesField10", FeatureType.CONTINUOUS),
        ("SalesField11", FeatureType.CONTINUOUS),
        ("SalesField12", FeatureType.CONTINUOUS),
        ("SalesField13", FeatureType.CONTINUOUS),
        ("SalesField14", FeatureType.CONTINUOUS),
        ("SalesField15", FeatureType.CONTINUOUS),
        ("PersonalField1", FeatureType.CONTINUOUS),
        ("PersonalField2", FeatureType.CONTINUOUS),
        ("PersonalField4A", FeatureType.CONTINUOUS),
        ("PersonalField4B", FeatureType.CONTINUOUS),
        ("PersonalField5", FeatureType.CONTINUOUS),
        ("PersonalField6", FeatureType.CONTINUOUS),
        ("PersonalField7", FeatureType.CATEGORICAL),
        ("PersonalField8", FeatureType.CONTINUOUS),
        ("PersonalField9", FeatureType.CONTINUOUS),
        ("PersonalField10A", FeatureType.CONTINUOUS),
        ("PersonalField10B", FeatureType.CONTINUOUS),
        ("PersonalField11", FeatureType.CONTINUOUS),
        ("PersonalField12", FeatureType.CONTINUOUS),
        ("PersonalField13", FeatureType.CONTINUOUS),
        ("PersonalField14", FeatureType.CONTINUOUS),
        ("PersonalField15", FeatureType.CONTINUOUS),
        ("PersonalField16", FeatureType.CATEGORICAL),
        ("PersonalField17", FeatureType.CATEGORICAL),
        ("PersonalField18", FeatureType.CATEGORICAL),
        ("PersonalField19", FeatureType.CATEGORICAL),
        ("PersonalField22", FeatureType.CONTINUOUS),
        ("PersonalField23", FeatureType.CONTINUOUS),
        ("PersonalField24", FeatureType.CONTINUOUS),
        ("PersonalField25", FeatureType.CONTINUOUS),
        ("PersonalField26", FeatureType.CONTINUOUS),
        ("PersonalField27", FeatureType.CONTINUOUS),
        ("PersonalField28", FeatureType.CONTINUOUS),
        ("PersonalField29", FeatureType.CONTINUOUS),
        ("PersonalField30", FeatureType.CONTINUOUS),
        ("PersonalField31", FeatureType.CONTINUOUS),
        ("PersonalField32", FeatureType.CONTINUOUS),
        ("PersonalField33", FeatureType.CONTINUOUS),
        ("PersonalField34", FeatureType.CONTINUOUS),
        ("PersonalField35", FeatureType.CONTINUOUS),
        ("PersonalField36", FeatureType.CONTINUOUS),
        ("PersonalField37", FeatureType.CONTINUOUS),
        ("PersonalField38", FeatureType.CONTINUOUS),
        ("PersonalField39", FeatureType.CONTINUOUS),
        ("PersonalField40", FeatureType.CONTINUOUS),
        ("PersonalField41", FeatureType.CONTINUOUS),
        ("PersonalField42", FeatureType.CONTINUOUS),
        ("PersonalField43", FeatureType.CONTINUOUS),
        ("PersonalField44", FeatureType.CONTINUOUS),
        ("PersonalField45", FeatureType.CONTINUOUS),
        ("PersonalField46", FeatureType.CONTINUOUS),
        ("PersonalField47", FeatureType.CONTINUOUS),
        ("PersonalField48", FeatureType.CONTINUOUS),
        ("PersonalField49", FeatureType.CONTINUOUS),
        ("PersonalField50", FeatureType.CONTINUOUS),
        ("PersonalField51", FeatureType.CONTINUOUS),
        ("PersonalField52", FeatureType.CONTINUOUS),
        ("PersonalField53", FeatureType.CONTINUOUS),
        ("PersonalField54", FeatureType.CONTINUOUS),
        ("PersonalField55", FeatureType.CONTINUOUS),
        ("PersonalField56", FeatureType.CONTINUOUS),
        ("PersonalField57", FeatureType.CONTINUOUS),
        ("PersonalField58", FeatureType.CONTINUOUS),
        ("PersonalField59", FeatureType.CONTINUOUS),
        ("PersonalField60", FeatureType.CONTINUOUS),
        ("PersonalField61", FeatureType.CONTINUOUS),
        ("PersonalField62", FeatureType.CONTINUOUS),
        ("PersonalField63", FeatureType.CONTINUOUS),
        ("PersonalField64", FeatureType.CONTINUOUS),
        ("PersonalField65", FeatureType.CONTINUOUS),
        ("PersonalField66", FeatureType.CONTINUOUS),
        ("PersonalField67", FeatureType.CONTINUOUS),
        ("PersonalField68", FeatureType.CONTINUOUS),
        ("PersonalField69", FeatureType.CONTINUOUS),
        ("PersonalField70", FeatureType.CONTINUOUS),
        ("PersonalField71", FeatureType.CONTINUOUS),
        ("PersonalField72", FeatureType.CONTINUOUS),
        ("PersonalField73", FeatureType.CONTINUOUS),
        ("PersonalField74", FeatureType.CONTINUOUS),
        ("PersonalField75", FeatureType.CONTINUOUS),
        ("PersonalField76", FeatureType.CONTINUOUS),
        ("PersonalField77", FeatureType.CONTINUOUS),
        ("PersonalField78", FeatureType.CONTINUOUS),
        ("PersonalField79", FeatureType.CONTINUOUS),
        ("PersonalField80", FeatureType.CONTINUOUS),
        ("PersonalField81", FeatureType.CONTINUOUS),
        ("PersonalField82", FeatureType.CONTINUOUS),
        ("PersonalField83", FeatureType.CONTINUOUS),
        ("PersonalField84", FeatureType.CATEGORICAL),
        ("PropertyField1A", FeatureType.CONTINUOUS),
        ("PropertyField1B", FeatureType.CONTINUOUS),
        ("PropertyField2A", FeatureType.CONTINUOUS),
        ("PropertyField2B", FeatureType.CONTINUOUS),
        ("PropertyField3", FeatureType.CATEGORICAL),
        ("PropertyField4", FeatureType.CATEGORICAL),
        ("PropertyField5", FeatureType.CATEGORICAL),
        ("PropertyField6", FeatureType.CONTINUOUS),
        ("PropertyField7", FeatureType.CATEGORICAL),
        ("PropertyField8", FeatureType.CONTINUOUS),
        ("PropertyField9", FeatureType.CONTINUOUS),
        ("PropertyField10", FeatureType.CONTINUOUS),
        ("PropertyField11A", FeatureType.CONTINUOUS),
        ("PropertyField11B", FeatureType.CONTINUOUS),
        ("PropertyField12", FeatureType.CONTINUOUS),
        ("PropertyField13", FeatureType.CONTINUOUS),
        ("PropertyField14", FeatureType.CATEGORICAL),
        ("PropertyField15", FeatureType.CONTINUOUS),
        ("PropertyField16A", FeatureType.CONTINUOUS),
        ("PropertyField16B", FeatureType.CONTINUOUS),
        ("PropertyField17", FeatureType.CONTINUOUS),
        ("PropertyField18", FeatureType.CONTINUOUS),
        ("PropertyField19", FeatureType.CONTINUOUS),
        ("PropertyField20", FeatureType.CONTINUOUS),
        ("PropertyField21A", FeatureType.CONTINUOUS),
        ("PropertyField21B", FeatureType.CONTINUOUS),
        ("PropertyField22", FeatureType.CONTINUOUS),
        ("PropertyField23", FeatureType.CONTINUOUS),
        ("PropertyField24A", FeatureType.CONTINUOUS),
        ("PropertyField24B", FeatureType.CONTINUOUS),
        ("PropertyField25", FeatureType.CONTINUOUS),
        ("PropertyField26A", FeatureType.CONTINUOUS),
        ("PropertyField26B", FeatureType.CONTINUOUS),
        ("PropertyField27", FeatureType.CONTINUOUS),
        ("PropertyField28", FeatureType.CATEGORICAL),
        ("PropertyField29", FeatureType.CATEGORICAL),
        ("PropertyField30", FeatureType.CATEGORICAL),
        ("PropertyField31", FeatureType.CATEGORICAL),
        ("PropertyField32", FeatureType.CATEGORICAL),
        ("PropertyField33", FeatureType.CATEGORICAL),
        ("PropertyField34", FeatureType.CATEGORICAL),
        ("PropertyField35", FeatureType.CONTINUOUS),
        ("PropertyField36", FeatureType.CATEGORICAL),
        ("PropertyField37", FeatureType.CATEGORICAL),
        ("PropertyField38", FeatureType.CATEGORICAL),
        ("PropertyField39A", FeatureType.CONTINUOUS),
        ("PropertyField39B", FeatureType.CONTINUOUS),
        ("GeographicField1A", FeatureType.CONTINUOUS),
        ("GeographicField1B", FeatureType.CONTINUOUS),
        ("GeographicField2A", FeatureType.CONTINUOUS),
        ("GeographicField2B", FeatureType.CONTINUOUS),
        ("GeographicField3A", FeatureType.CONTINUOUS),
        ("GeographicField3B", FeatureType.CONTINUOUS),
        ("GeographicField4A", FeatureType.CONTINUOUS),
        ("GeographicField4B", FeatureType.CONTINUOUS),
        ("GeographicField5A", FeatureType.CONTINUOUS),
        ("GeographicField5B", FeatureType.CONTINUOUS),
        ("GeographicField6A", FeatureType.CONTINUOUS),
        ("GeographicField6B", FeatureType.CONTINUOUS),
        ("GeographicField7A", FeatureType.CONTINUOUS),
        ("GeographicField7B", FeatureType.CONTINUOUS),
        ("GeographicField8A", FeatureType.CONTINUOUS),
        ("GeographicField8B", FeatureType.CONTINUOUS),
        ("GeographicField9A", FeatureType.CONTINUOUS),
        ("GeographicField9B", FeatureType.CONTINUOUS),
        ("GeographicField10A", FeatureType.CONTINUOUS),
        ("GeographicField10B", FeatureType.CONTINUOUS),
        ("GeographicField11A", FeatureType.CONTINUOUS),
        ("GeographicField11B", FeatureType.CONTINUOUS),
        ("GeographicField12A", FeatureType.CONTINUOUS),
        ("GeographicField12B", FeatureType.CONTINUOUS),
        ("GeographicField13A", FeatureType.CONTINUOUS),
        ("GeographicField13B", FeatureType.CONTINUOUS),
        ("GeographicField14A", FeatureType.CONTINUOUS),
        ("GeographicField14B", FeatureType.CONTINUOUS),
        ("GeographicField15A", FeatureType.CONTINUOUS),
        ("GeographicField15B", FeatureType.CONTINUOUS),
        ("GeographicField16A", FeatureType.CONTINUOUS),
        ("GeographicField16B", FeatureType.CONTINUOUS),
        ("GeographicField17A", FeatureType.CONTINUOUS),
        ("GeographicField17B", FeatureType.CONTINUOUS),
        ("GeographicField18A", FeatureType.CONTINUOUS),
        ("GeographicField18B", FeatureType.CONTINUOUS),
        ("GeographicField19A", FeatureType.CONTINUOUS),
        ("GeographicField19B", FeatureType.CONTINUOUS),
        ("GeographicField20A", FeatureType.CONTINUOUS),
        ("GeographicField20B", FeatureType.CONTINUOUS),
        ("GeographicField21A", FeatureType.CONTINUOUS),
        ("GeographicField21B", FeatureType.CONTINUOUS),
        ("GeographicField22A", FeatureType.CONTINUOUS),
        ("GeographicField22B", FeatureType.CONTINUOUS),
        ("GeographicField23A", FeatureType.CONTINUOUS),
        ("GeographicField23B", FeatureType.CONTINUOUS),
        ("GeographicField24A", FeatureType.CONTINUOUS),
        ("GeographicField24B", FeatureType.CONTINUOUS),
        ("GeographicField25A", FeatureType.CONTINUOUS),
        ("GeographicField25B", FeatureType.CONTINUOUS),
        ("GeographicField26A", FeatureType.CONTINUOUS),
        ("GeographicField26B", FeatureType.CONTINUOUS),
        ("GeographicField27A", FeatureType.CONTINUOUS),
        ("GeographicField27B", FeatureType.CONTINUOUS),
        ("GeographicField28A", FeatureType.CONTINUOUS),
        ("GeographicField28B", FeatureType.CONTINUOUS),
        ("GeographicField29A", FeatureType.CONTINUOUS),
        ("GeographicField29B", FeatureType.CONTINUOUS),
        ("GeographicField30A", FeatureType.CONTINUOUS),
        ("GeographicField30B", FeatureType.CONTINUOUS),
        ("GeographicField31A", FeatureType.CONTINUOUS),
        ("GeographicField31B", FeatureType.CONTINUOUS),
        ("GeographicField32A", FeatureType.CONTINUOUS),
        ("GeographicField32B", FeatureType.CONTINUOUS),
        ("GeographicField33A", FeatureType.CONTINUOUS),
        ("GeographicField33B", FeatureType.CONTINUOUS),
        ("GeographicField34A", FeatureType.CONTINUOUS),
        ("GeographicField34B", FeatureType.CONTINUOUS),
        ("GeographicField35A", FeatureType.CONTINUOUS),
        ("GeographicField35B", FeatureType.CONTINUOUS),
        ("GeographicField36A", FeatureType.CONTINUOUS),
        ("GeographicField36B", FeatureType.CONTINUOUS),
        ("GeographicField37A", FeatureType.CONTINUOUS),
        ("GeographicField37B", FeatureType.CONTINUOUS),
        ("GeographicField38A", FeatureType.CONTINUOUS),
        ("GeographicField38B", FeatureType.CONTINUOUS),
        ("GeographicField39A", FeatureType.CONTINUOUS),
        ("GeographicField39B", FeatureType.CONTINUOUS),
        ("GeographicField40A", FeatureType.CONTINUOUS),
        ("GeographicField40B", FeatureType.CONTINUOUS),
        ("GeographicField41A", FeatureType.CONTINUOUS),
        ("GeographicField41B", FeatureType.CONTINUOUS),
        ("GeographicField42A", FeatureType.CONTINUOUS),
        ("GeographicField42B", FeatureType.CONTINUOUS),
        ("GeographicField43A", FeatureType.CONTINUOUS),
        ("GeographicField43B", FeatureType.CONTINUOUS),
        ("GeographicField44A", FeatureType.CONTINUOUS),
        ("GeographicField44B", FeatureType.CONTINUOUS),
        ("GeographicField45A", FeatureType.CONTINUOUS),
        ("GeographicField45B", FeatureType.CONTINUOUS),
        ("GeographicField46A", FeatureType.CONTINUOUS),
        ("GeographicField46B", FeatureType.CONTINUOUS),
        ("GeographicField47A", FeatureType.CONTINUOUS),
        ("GeographicField47B", FeatureType.CONTINUOUS),
        ("GeographicField48A", FeatureType.CONTINUOUS),
        ("GeographicField48B", FeatureType.CONTINUOUS),
        ("GeographicField49A", FeatureType.CONTINUOUS),
        ("GeographicField49B", FeatureType.CONTINUOUS),
        ("GeographicField50A", FeatureType.CONTINUOUS),
        ("GeographicField50B", FeatureType.CONTINUOUS),
        ("GeographicField51A", FeatureType.CONTINUOUS),
        ("GeographicField51B", FeatureType.CONTINUOUS),
        ("GeographicField52A", FeatureType.CONTINUOUS),
        ("GeographicField52B", FeatureType.CONTINUOUS),
        ("GeographicField53A", FeatureType.CONTINUOUS),
        ("GeographicField53B", FeatureType.CONTINUOUS),
        ("GeographicField54A", FeatureType.CONTINUOUS),
        ("GeographicField54B", FeatureType.CONTINUOUS),
        ("GeographicField55A", FeatureType.CONTINUOUS),
        ("GeographicField55B", FeatureType.CONTINUOUS),
        ("GeographicField56A", FeatureType.CONTINUOUS),
        ("GeographicField56B", FeatureType.CONTINUOUS),
        ("GeographicField57A", FeatureType.CONTINUOUS),
        ("GeographicField57B", FeatureType.CONTINUOUS),
        ("GeographicField58A", FeatureType.CONTINUOUS),
        ("GeographicField58B", FeatureType.CONTINUOUS),
        ("GeographicField59A", FeatureType.CONTINUOUS),
        ("GeographicField59B", FeatureType.CONTINUOUS),
        ("GeographicField60A", FeatureType.CONTINUOUS),
        ("GeographicField60B", FeatureType.CONTINUOUS),
        ("GeographicField61A", FeatureType.CONTINUOUS),
        ("GeographicField61B", FeatureType.CONTINUOUS),
        ("GeographicField62A", FeatureType.CONTINUOUS),
        ("GeographicField62B", FeatureType.CONTINUOUS),
        ("GeographicField63", FeatureType.CATEGORICAL),
        ("GeographicField64", FeatureType.CATEGORICAL),
    ])



"""Module for generating auto-ml data.

TODO: move this to its own project at top level `ml-research` root

This module should provide an interface to generate the following training data

create a set of utility functions that creates the following tuples:
(is_executable, creates_estimator, creates_pipeline, code_string)
"""


import inspect
import numpy as np
import random
import re
import sklearn
import string
import warnings

from collections import namedtuple
from sklearn.utils.testing import all_estimators


CHARACTERS = string.ascii_letters + string.digits + "._()"
# These estimators have been deprecated
EXCLUDE_ESTIMATORS = [
    "RandomizedLogisticRegression",
    "RandomizedLasso",
    "RandomizedPCA",
    "DictionaryLearning",
    "MiniBatchDictionaryLearning",
    "GaussianProcess",
]
EXECUTABLE_MAP = {
    True: "executable",
    False: "not_executable",
}
CREATES_ESTIMATOR = {
    True: "creates_estimator",
    False: "not_creates_estimator",
}


HyperparameterSpec = namedtuple(
    "HyperparameterSpec", ["hyperparameters", "defaults"])


class AlgorithmEnv(object):

    def __init__(self, classifiers, regressors, transformers):
        """Initialize Algorithm and Hyperparameter Space (AHS).

        The AlgorithmEnv is the environment from which we sample algorithms,
        which are defined in this context as sklearn Transformer and Estimator
        objects.

        :param list[AlgorithmSpec] classifiers: list of classifiers to search
        :param list[AlgorithmSpec] regressors: list of regressors to search
        :param list[AlgorithmSpec] transformers: list of transformers to search
        """
        self.classifiers = classifiers
        self.regressors = regressors
        self.transformers = transformers

    def sample_algorithm_code(self, type_filter=None):
        """Samples an algorithm."""
        sample_code = self._sample_algorithm_code(type_filter)
        return self._input_data(sample_code)

    def algorithm_obj_to_instance(self, sample):
        """Convert algorithm code so that code evaluates to instance."""
        _, sample_code = sample
        return self._input_data("%s()" % sample_code)

    def mutate_sample(self, sample, mutate_all=False):
        """Mutates a sample.

        :param tuple sample: (metafeatures <list>, sample_code <str>)
        :returns: mutated sample
        """
        _, sample_code = sample
        return self._input_data(
            self._mutate_sample_code(sample_code, mutate_all))

    def _sample_algorithm_code(self, type_filter):
        """Samples an algorithm from the algorithm environment.

        :param str type_filter: {"classifiers", "regressors", "transformers"}
        :returns BaseEstimator: an sklearn algorithm (estimator/transformer)
        """
        algorithm_sample_space = None
        if type_filter is None:
            algorithm_sample_space = self.classifiers + self.regressors + \
                self.transformers
        elif type_filter == "classifiers":
            algorithm_sample_space = self.classifiers
        elif type_filter == "regressors":
            algorithm_sample_space = self.regressors
        elif type_filter == "transformers":
            algorithm_sample_space = self.transformers
        else:
            raise ValueError(
                "unrecognized type_filter argument: %s" % type_filter)
        return random.choice(algorithm_sample_space).estimator_code

    def _mutate_sample_code(self, sample_code, mutate_all):
        """Randomly mutate a random number of characters in the code."""
        code_length = len(sample_code)
        mutate_n = len(code_length) if mutate_all else \
            random.randint(0, code_length)
        mutate_indices = np.random.choice(
            range(code_length), mutate_n, replace=False)
        sample_code_split = [s for s in sample_code]
        for i in mutate_indices:
            char = CHARACTERS[random.randint(0, len(CHARACTERS) - 1)]
            sample_code_split[i] = char
        return "".join(sample_code_split)

    def _input_data(self, sample_code):
        """Creates a data sample in the form (metafeatures, sample_code)."""
        is_executable = code_is_executable(sample_code)
        evals_to_algorithm_instance = False
        if is_executable:
            evals_to_algorithm_instance = code_evals_to_algorithm_instance(
                sample_code)
        metafeatures = [
            EXECUTABLE_MAP[is_executable],
            CREATES_ESTIMATOR[evals_to_algorithm_instance]]
        return (metafeatures, sample_code)


class AlgorithmSpec(object):

    def __init__(self, estimator_name, estimator_class):
        self.estimator_name = estimator_name
        self.estimator_class = estimator_class

    @property
    def hyperparameters(self):
        return self._get_hyperparams(self.estimator_class)

    @property
    def estimator_code(self):
        return self._get_estimator_string(self.estimator_class)

    def _get_hyperparams(self, estimator_class):
        """Get hyperparameters and their defaults from an estimator class.

        :param BaseEstimator estimator_class: sklearn estimator or transformer
            class.
        :returns HyperparameterSpec: a specification of the
            hyperparameter space, with attributes:
            - hyperparameters: list[str]
            - default: tuple[str|int|float|None|...]
        """
        argspec = inspect.getfullargspec(estimator_class)
        hyperparams = [arg for arg in argspec.args if arg != "self"]
        if len(hyperparams) > 0 and argspec.defaults is not None:
            assert len(hyperparams) == len(argspec.defaults)
        return HyperparameterSpec(hyperparams, argspec.defaults)

    def _get_estimator_string(self, estimator_class):
        """Extract string representation from Estimator class."""
        return re.search(
            "<class '(.+)'>", str(estimator_class)).groups()[0]

    def __repr__(self):
        return "AlgorithmSpec__%s" % self.estimator_name



def create_algorithm_env():
    """Create Algorithm Environment, used to sample algorithms, hyperparameters

    TODO: This should probably implemented as a class so that methods can be
    better organized around this concept.

    :returns AlgorithmEnv: an object containing dictionaries specifying the
        search space of the environment.
    """
    return AlgorithmEnv(
        get_algorithms("classifier"),
        get_algorithms("regressor"),
        get_algorithms("transformer"))


def get_algorithms(estimator_type):
    """Get estimators of a particular type.

    This function relies on the ``all_estimators`` function in
    sklearn.utils.testing.

    :param str estimator_type: {'classifier', 'regressor', 'transformer'}
        specifies the type of estimator to get.
    :returns dict[str, object]: where keys are the name of the estimator and
        values are the corresponding class.
    """
    with warnings.catch_warnings():
        # catch deprecation warning for modules to be removed in 0.20 in favor
        # of the model_selection module: cross_validation, grid_search, and
        # learning_curve
        warnings.simplefilter("ignore")
        return [
            AlgorithmSpec(estimator_name, estimator_class) for
            estimator_name, estimator_class in
            all_estimators(type_filter=estimator_type)
            if estimator_name not in EXCLUDE_ESTIMATORS]


def code_is_executable(estimator_code):
    """Determine if code string is executable.

    :param str estimator_code: string to check.
    :returns bool: True if executable
    """
    try:
        eval(estimator_code)
        return True
    except:
        return False


def code_evals_to_algorithm_instance(estimator_code):
    """Determing if code evaluates to a BaseEstimator instance."""
    parent_classes = inspect.getmro(type(eval(estimator_code)))
    return sklearn.base.BaseEstimator in parent_classes


if __name__ == "__main__":
    algorithm_env = create_algorithm_env()
    print("Algorithm Environment:")
    print("classifiers: %s" % len(algorithm_env.classifiers))
    print("regressors: %s" % len(algorithm_env.regressors))
    print("transformers: %s" % len(algorithm_env.transformers))

    print("\nGenerating sample data from environment:")
    for _ in range(5):
        sample_data = algorithm_env.sample_algorithm_code()
        print(sample_data)
        print(algorithm_env.algorithm_obj_to_instance(sample_data))
        print(algorithm_env.mutate_sample(sample_data))
        print(algorithm_env.mutate_sample(sample_data, mutate_all=False))
        print("")



"""Reinforce module for training the CASH controller."""

import numpy as np
import torch
import torch.nn as nn

from collections import defaultdict
from torch.autograd import Variable

from . import utils, loggers
from .tracking import MetricsTracker

SINGLE_BASELINE = "all_data_envs"
EPSILON = np.finfo(np.float32).eps.item()


class MetaLearnReinforce(object):
    """Reinforce component of deep-cash algorithm."""

    def __init__(
            self,
            controller,
            t_env,
            beta=0.99,
            entropy_coef=0.0,
            with_baseline=True,
            single_baseline=True,
            normalize_reward=True,
            metrics_logger=loggers.default_logger):
        """Initialize CASH Reinforce Algorithm.

        :param pytorch.nn.Module controller: A CASH controller to select
            actions.
        :param TaskEnvironment t_env: task environment to sample data
            environments and evaluate proposed ml frameworks.
        :param float beta: hyperparameter for exponential moving average to
            compute baseline reward (used to regularize REINFORCE).
        :param float entropy_coef: coefficient for entropy regularization.
        :param bool with_baseline: whether or not to regularize the controller
            with the exponential moving average of past rewards.
        :param bool single_baseline: if True, maintains a single baseline
            reward buffer, otherwise maintain a separate baseline buffer for
            each data environment available to the task environment.
        :param bool normalize_reward: whether or not to mean-center and
            standard-deviation-scale the reward signal for backprop.
        :param callable metrics_logger: logging function to use. The function
            takes as input a tracking.TrackerBase object and prints out a
            message.
        :ivar defaultdict[str -> list] _baseline_fn: a map that keeps track
            of the baseline function for a particular task environment state
            (i.e. the sampled data environment at a particular episode).
        """
        self.controller = controller
        self.t_env = t_env
        self._beta = beta
        self._entropy_coef = entropy_coef
        self._with_baseline = with_baseline
        self._single_baseline = single_baseline
        self._normalize_reward = normalize_reward
        self._metrics_logger = metrics_logger

    def fit(self, optim, optim_kwargs, n_episodes=100, n_iter=100,
            verbose=True, procnum=None):
        """Fits the CASH controller with the REINFORCE algorithm.

        :param torch.optim optim: type of optimization to use for backprop.
        :param dict optim_kwargs: arguments to pass to optim constructor
        :param int n_episodes: number of episodes to train.
        :param int n_iter: number of iterations per episode.
        :param bool verbose: whether or not to print the exponential mean
            reward per iteration.
        :param int procnum: optional argument to indicate the process number
            used for multiprocessing.
        """
        # set optimization object
        self.optim = optim(self.controller.parameters(), **optim_kwargs)
        self._n_episodes = n_episodes

        # track metrics
        self.tracker = MetricsTracker()

        # baseline reward
        self._baseline_fn = {
            "buffer": defaultdict(list),
            "current": defaultdict(int),
        }
        # the buffer history keeps track of baseline rewards across all the
        # episodes, used for record-keeping and testing purposes.
        self._baseline_buffer_history = defaultdict(list)

        for i_episode in range(self._n_episodes):
            self.t_env.sample_data_env()
            self._action_buffer = []
            self._validation_scores = []
            self._successful_mlfs = []
            self._algorithm_sets = []
            self._hyperparameter_sets = []
            self._best_validation_score = None
            self._best_mlf = None
            msg = "episode %d, task: %s" % (
                i_episode, self.t_env.current_data_env.name)
            if procnum is not None:
                msg = "proc num: %d, %s" % (procnum, msg)
            print("\n" + msg)
            # since the action sequences within episode are no longer
            # independent, need to get a single baseline value (the
            # exponential mean of the return for the previous episode)
            buffer = self._baseline_buffer()
            if len(buffer) == 0:
                baseline = self._baseline_current()
            else:
                # get last baseline value
                baseline = buffer[-1]
                del self._baseline_buffer()[:]

            self.start_episode(n_iter, verbose)

            # accumulate stats
            n_successful_mlfs = len(self._successful_mlfs)
            n_unique_mlfs = len(set((tuple(s) for s in self._algorithm_sets)))
            n_unique_hyperparams = len(set(
                [str(d.items()) for d in self._hyperparameter_sets]))
            mlf_diversity = utils._diversity_metric(
                n_unique_mlfs, n_successful_mlfs)
            hyperparam_diversity = utils._diversity_metric(
                n_unique_hyperparams, n_successful_mlfs)

            mean_rewards = np.mean(self.controller.reward_buffer)
            if len(self._validation_scores) > 0:
                mean_val_scores = np.mean(self._validation_scores)
                std_val_scores = np.std(self._validation_scores)
            else:
                mean_val_scores, std_val_scores = np.nan, np.nan

            # backward pass through controller
            loss, grad_agg = self.end_episode(baseline, mlf_diversity)

            self.tracker.update_metrics({
                "episode": i_episode + 1,
                "data_env_names": self.t_env.current_data_env.name,
                "scorers": self.t_env.scorer.name,
                "losses": loss,
                "mean_rewards": mean_rewards,
                "aggregate_gradients": grad_agg,
                "mean_validation_scores": mean_val_scores,
                "std_validation_scores": std_val_scores,
                "best_validation_scores": self._best_validation_score,
                "best_mlfs": self._best_mlf,
                "n_successful_mlfs": n_successful_mlfs,
                "n_unique_mlfs": n_unique_mlfs,
                "n_unique_hyperparams": n_unique_hyperparams,
                "mlf_diversity": mlf_diversity,
                "hyperparam_diversity": hyperparam_diversity
            })
            if self._metrics_logger is not None:
                self._metrics_logger(self.tracker)
        return self

    @property
    def history(self):
        """Get metrics history per episode."""
        return self.tracker.history

    @property
    def best_mlfs(self):
        """Get best mlfs per episode."""
        return self.tracker.best_mlfs

    def start_episode(self, n_iter, verbose):
        """Begin training the controller."""
        self._n_valid_mlf = 0
        prev_reward, prev_action = 0, self.controller.init_action()
        for i in range(n_iter):
            prev_reward, prev_action = self._fit_iter(
                self.t_env.sample_task_state(),
                self.t_env.current_data_env.target_type,
                prev_reward,
                prev_action)
            if verbose:
                print(
                    "iter %d - n valid mlf: %d/%d" % (
                        i, self._n_valid_mlf, i + 1),
                    sep=" ", end="\r", flush=True)

    def _fit_iter(
            self, metafeature_tensor, target_type, prev_reward, prev_action):
        actions, action_activation = self.controller.decode(
            init_input_tensor=prev_action,
            target_type=target_type,
            aux=utils.aux_tensor(prev_reward),
            metafeatures=Variable(metafeature_tensor),
            init_hidden=self.controller.init_hidden())
        reward = self.evaluate_actions(actions, action_activation)
        # TODO: directly update the underlying data structures
        self._update_log_prob_buffer(actions)
        self._update_reward_buffer(reward)
        self._update_entropy_buffer(actions)
        self._update_baseline_reward_buffer(reward)
        for k, v in self._baseline_fn["buffer"].items():
            self._baseline_buffer_history[k].extend(v)
        return reward, Variable(action_activation.data)

    def evaluate_actions(self, actions, action_activation):
        """Evaluate actions on the validation set of the data environment."""
        algorithms, hyperparameters = utils.get_mlf_components(actions)
        mlf = self.controller.a_space.create_ml_framework(
            algorithms, hyperparameters=hyperparameters,
            task_metadata=self.t_env.get_current_task_metadata())
        mlf, reward, score = self.t_env.evaluate(mlf)
        self._action_buffer.append(action_activation)
        self._algorithm_sets.append(algorithms)
        self._hyperparameter_sets.append(hyperparameters)
        if reward is None:
            return self.t_env.error_reward
        else:
            self._validation_scores.append(score)
            self._n_valid_mlf += 1
            self._successful_mlfs.append(mlf)
            if self._best_validation_score is None or \
                    self.t_env.scorer.comparator(
                        score, self._best_validation_score):
                self._best_validation_score = score
                self._best_mlf = mlf
            return reward

    def end_episode(self, baseline, mlf_diversity):
        """End an episode with one backpropagation step.

        Since REINFORCE algorithm is a gradient ascent method, negate the log
        probability in order to do gradient descent on the negative expected
        rewards.

        :param float baseline: baseline reward value, the exponential mean of
            rewards over previous episodes.
        :param float mlf_diversity: 1.0 means all successfully fitted mlfs
            are unique, 0.0 if there is only a single unique mlf in the batch.
        :returns: tuple of loss and aggregated gradient.
        :rtype: tuple[float, float]
        """
        _check_buffers(
            self.controller.log_prob_buffer,
            self.controller.reward_buffer,
            self.controller.entropy_buffer)
        self.optim.zero_grad()
        n = len(self.controller.reward_buffer)

        R = torch.FloatTensor(self.controller.reward_buffer)
        if self._with_baseline:
            R = R - baseline
        if self._normalize_reward:
            R = normalize_reward(R)

        # compute loss
        loss = [-l * r - self._entropy_coef * e
                for log_probs, r, entropies in zip(
                    self.controller.log_prob_buffer, R,
                    self.controller.entropy_buffer)
                for l, e in zip(log_probs, entropies)]

        loss = torch.cat(loss).sum().div(n)

        # one step of gradient descent
        loss.backward()
        # gradient clipping to prevent exploding gradient
        nn.utils.clip_grad_norm_(self.controller.parameters(), 20)
        self.optim.step()

        grad_agg = sum([
            p.grad.data.sum() for p in self.controller.parameters()
            if p.grad is not None])

        # reset rewards and log probs
        del self.controller.log_prob_buffer[:]
        del self.controller.reward_buffer[:]
        del self.controller.entropy_buffer[:]

        return loss.data.item(), grad_agg

    def _update_log_prob_buffer(self, actions):
        self.controller.log_prob_buffer.append(
            [a["log_prob"] for a in actions])

    def _update_reward_buffer(self, reward):
        self.controller.reward_buffer.append(reward)

    def _update_entropy_buffer(self, actions):
        self.controller.entropy_buffer.append([a["entropy"] for a in actions])

    def _update_baseline_reward_buffer(self, reward):
        buffer = self._baseline_buffer()
        current = self._baseline_current()
        buffer.append(current)
        self._update_current_baseline(
            utils._exponential_mean(reward, current, beta=self._beta))

    def _baseline_fn_key(self):
        return SINGLE_BASELINE if self._single_baseline else \
            self.t_env.current_data_env.name

    def _baseline_buffer(self):
        return self._baseline_fn["buffer"][self._baseline_fn_key()]

    def _baseline_current(self):
        return self._baseline_fn["current"][self._baseline_fn_key()]

    def _update_current_baseline(self, new_baseline):
        self._baseline_fn["current"][self._baseline_fn_key()] = new_baseline


def _check_buffers(log_prob_buffer, reward_buffer, entropy_buffer):
    n_abuf = len(log_prob_buffer)
    n_rbuf = len(reward_buffer)
    n_ebuf = len(entropy_buffer)
    if n_abuf != n_rbuf != n_ebuf:
        raise ValueError(
            "all buffers need the same length, found:\n"
            "log_prob_buffer = %d\n"
            "reward_buffer = %d\n"
            "entropy_buffer = %d\n" % (n_abuf, n_rbuf, n_ebuf))


def normalize_reward(rbuffer):
    """Mean-center and std-rescale the reward buffer.

    :param torch.FloatTensor rbuffer: list of rewards from an episode
    :returns: normalized tensor
    """
    return (rbuffer - rbuffer.mean()).div(rbuffer.std() + EPSILON)


def print_actions(actions):
    """Pretty print actions for an ml framework."""
    for i, action in enumerate(actions):
        print("action #: %s" % i)
        for k, v in action.items():
            print("%s: %s" % (k, v))



"""Module for creating, reading deep cash experiments."""

import datetime
import inspect
import logging
import os
import pandas as pd
import subprocess
import torch
import torch.multiprocessing as mp
import yaml
import yamlordereddictloader

from collections import namedtuple, OrderedDict
from functools import partial
from pathlib import Path
from shutil import rmtree
from sklearn.externals import joblib

from .algorithm_space import AlgorithmSpace
from .task_environment import TaskEnvironment
from .metalearn_controller import MetaLearnController
from .metalearn_reinforce import MetaLearnReinforce
from .data_environments.environments import envs
from .data_types import ExperimentType
from .loggers import get_loggers, empty_logger
from .random_search import CASHRandomSearch
from . import utils

# TODO: support ability to copy a config file with new name
# TODO: support ability to run floyd job by experiment name


logger = logging.getLogger(__name__)


ExperimentConfig = namedtuple(
    "ExperimentConfig", [
        "name",
        "experiment_type",
        "description",
        "created_at",
        "git_hash",
        "parameters"])


def _env_names():
    return [d["dataset_name"] for d in envs()]


def get_experiment_fn(experiment_type):
    return {
        ExperimentType.METALEARN_REINFORCE: run_experiment,
        ExperimentType.RANDOM_SEARCH: run_random_search,
    }[ExperimentType[experiment_type]]


def get_default_parameters(experiment_type):
    experiment_fn = get_experiment_fn(experiment_type)
    return OrderedDict([
        (k, v.default) for k, v in
        inspect.signature(experiment_fn).parameters.items()
    ])


def create_config(name, experiment_type, description,
                  date_format="%Y-%m-%d-%H-%M-%S", **custom_parameters):
    parameters = get_default_parameters(experiment_type)
    if custom_parameters:
        parameters.update(custom_parameters)
    created_at = datetime.datetime.now()
    git_hash = subprocess.check_output(["git", "rev-parse", "--short", "HEAD"])
    return ExperimentConfig(
        name=name,
        experiment_type=experiment_type,
        description=description,
        created_at=created_at.strftime(date_format),
        git_hash=git_hash.strip().decode("utf-8"),
        parameters=parameters)


def write_config(config, dir_path):
    dir_path = Path(dir_path)
    dir_path.mkdir(exist_ok=True)
    fp = _config_filepath(dir_path, config)
    with open(str(fp), "w") as f:
        yaml.dump(config._asdict(), f, Dumper=yamlordereddictloader.Dumper,
                  default_flow_style=False)
    print("wrote experiment config file to %s" % fp)


def read_config(fp):
    with open(fp, "r") as f:
        return yaml.load(f, Loader=yamlordereddictloader.Loader)


def _config_filepath(dir_path, config):
    return dir_path / (
        "experiment_%s_%s.yml" % (config.created_at, config.name))


def gather_history(return_dict):
    history = []
    for i, h in return_dict.items():
        h = pd.DataFrame(h).assign(trial_number=i)
        history.append(h)
    return pd.concat(history)


def save_best_mlfs(data_path, best_mlfs, procnum):
    mlf_path = data_path / ("metalearn_controller_mlfs_trial_%d" % procnum)
    mlf_path.mkdir(exist_ok=True)
    for i, mlf in enumerate(best_mlfs):
        joblib.dump(mlf, mlf_path / ("best_mlf_episode_%d.pkl" % (i + 1)))


def save_experiment(history, data_path, fname):
    # save history of all trials
    history.to_csv(str(data_path / fname), index=False)


def run_experiment(
        datasets=None,
        output_fp=os.path.dirname(__file__) + "/../output",
        n_trials=1,
        input_size=30,
        hidden_size=30,
        output_size=30,
        n_layers=3,
        dropout_rate=0.2,
        beta=0.9,
        entropy_coef=0.0,
        with_baseline=True,
        single_baseline=True,
        normalize_reward=False,
        n_episodes=100,
        n_iter=16,
        learning_rate=0.005,
        env_sources=["SKLEARN", "OPEN_ML", "KAGGLE"],
        target_types=["BINARY", "MULTICLASS"],
        test_set_config=OrderedDict([
            ("SKLEARN", OrderedDict([
                ("test_size", 0.8),
                ("random_state", 100)]))
        ]),
        error_reward=0,
        n_samples=5000,
        per_framework_time_limit=180,
        per_framework_memory_limit=5000,
        metric_logger=None,
        fit_verbose=1,
        controller_seed=1000,
        task_environment_seed=100):
    """Run deep cash experiment with single configuration."""
    print("Running cash controller experiment with %d %s" % (
        n_trials, "trials" if n_trials > 1 else "trial"))
    torch.manual_seed(controller_seed)
    output_fp = os.path.dirname(__file__) + "/../output" if \
        output_fp is None else output_fp
    data_path = Path(output_fp)

    # initialize error logging (this is to log fit/predict/score errors made
    # when evaluating a proposed MLF)
    utils.init_logging(str(Path(output_fp) / "fit_predict_error_logs.log"))

    # this logger is for logging metrics to floydhub/stdout
    metric_logger = get_loggers().get(metric_logger, empty_logger)

    def worker(procnum, reinforce, return_dict):
        """Fit REINFORCE Helper function."""
        reinforce.fit(
            optim=torch.optim.Adam,
            optim_kwargs={"lr": learning_rate},
            n_episodes=n_episodes,
            n_iter=n_iter,
            verbose=bool(int(fit_verbose)),
            procnum=procnum)
        # serialize reinforce controller here
        reinforce.controller.save(data_path / ("controller_trial_%d.pt" % i))
        # serialize best mlfs
        save_best_mlfs(data_path, reinforce.best_mlfs, procnum)
        return_dict[procnum] = reinforce.history

    # multiprocessing manager
    manager = mp.Manager()
    return_dict = manager.dict()
    processes = []

    for i in range(n_trials):
        t_env = TaskEnvironment(
            env_sources=env_sources,
            target_types=target_types,
            test_set_config=test_set_config,
            random_state=task_environment_seed,
            per_framework_time_limit=per_framework_time_limit,
            per_framework_memory_limit=per_framework_memory_limit,
            dataset_names=datasets,
            error_reward=error_reward,
            n_samples=n_samples)

        # create algorithm space
        a_space = AlgorithmSpace(
            with_end_token=False,
            hyperparam_with_start_token=False,
            hyperparam_with_none_token=False)

        controller = MetaLearnController(
            metafeature_size=t_env.metafeature_dim,
            input_size=input_size,
            hidden_size=hidden_size,
            output_size=output_size,
            a_space=a_space,
            dropout_rate=dropout_rate,
            num_rnn_layers=n_layers)

        reinforce = MetaLearnReinforce(
            controller, t_env,
            beta=beta,
            entropy_coef=entropy_coef,
            with_baseline=with_baseline,
            single_baseline=single_baseline,
            normalize_reward=normalize_reward,
            metrics_logger=partial(metric_logger, prefix="proc_num_%d" % i))

        p = mp.Process(target=worker, args=(i, reinforce, return_dict))
        p.start()
        processes.append(p)

    for p in processes:
        p.join()

    save_experiment(
        gather_history(return_dict),
        data_path, "rnn_metalearn_controller_experiment.csv")


def run_random_search(
        datasets=None,
        output_fp=os.path.dirname(__file__) + "/../output",
        n_trials=1,
        n_episodes=100,
        n_iter=16,
        env_sources=["SKLEARN", "OPEN_ML", "KAGGLE"],
        target_types=["BINARY", "MULTICLASS"],
        test_set_config=OrderedDict([
            ("SKLEARN", OrderedDict([
                ("test_size", 0.8),
                ("random_state", 100)]))
        ]),
        error_reward=0,
        n_samples=5000,
        per_framework_time_limit=180,
        per_framework_memory_limit=5000,
        metric_logger="default",
        fit_verbose=1,
        controller_seed=1000,
        task_environment_seed=100):
    """Run deep cash experiment with single configuration."""
    print("Running random search experiment with %d %s" % (
        n_trials, "trials" if n_trials > 1 else "trial"))
    output_fp = os.path.dirname(__file__) + "/../output" if \
        output_fp is None else output_fp
    data_path = Path(output_fp)

    # initialize error logging (this is to log fit/predict/score errors made
    # when evaluating a proposed MLF)
    utils.init_logging(
        str(Path(output_fp) / "fit_predict_error_logs_randomsearch.log"))

    # this logger is for logging metrics to floydhub/stdout
    metric_logger = get_loggers().get(metric_logger, empty_logger)

    def worker(procnum, random_search, return_dict):
        """Fit REINFORCE Helper function."""
        random_search.fit(
            n_episodes=n_episodes,
            n_iter=n_iter,
            verbose=fit_verbose,
            procnum=procnum)
        save_best_mlfs(data_path, random_search.best_mlfs, procnum)
        return_dict[procnum] = random_search.history

    # multiprocessing manager
    manager = mp.Manager()
    return_dict = manager.dict()
    processes = []

    for i in range(n_trials):
        t_env = TaskEnvironment(
            env_sources=env_sources,
            target_types=target_types,
            test_set_config=test_set_config,
            random_state=task_environment_seed,
            per_framework_time_limit=per_framework_time_limit,
            per_framework_memory_limit=per_framework_memory_limit,
            dataset_names=datasets,
            error_reward=error_reward,
            n_samples=n_samples)

        # create algorithm space
        a_space = AlgorithmSpace(
            with_end_token=False,
            hyperparam_with_start_token=False,
            hyperparam_with_none_token=False)

        random_search = CASHRandomSearch(
            a_space=a_space,
            t_env=t_env,
            metrics_logger=partial(metric_logger, prefix="proc_num_%d" % i))

        p = mp.Process(target=worker, args=(i, random_search, return_dict))
        p.start()
        processes.append(p)

    for p in processes:
        p.join()

    save_experiment(
        gather_history(return_dict),
        data_path, "rnn_cash_randomsearch_experiment.csv")



"""An RNN to generate sklearn code.

TODO: move this to its own project at top level `ml-research` root

IDEAS:
- instead of sampling the max probability prediction, sample based on the
  probability distribution of the softmax.
- use skip connections i.e. resnets.
- use an attention model to generate code.
"""

import math
import time
import torch
import torch.nn as nn
from torch.autograd import Variable

from .algorithm_env import CHARACTERS


metafeature_categories = (
    ["not_executable", "executable"],
    ["not_creates_estimator", "creates_estimator"],
)
n_metafeatures = sum([len(m) for m in metafeature_categories])
sos_token = "^"  # start of sentence token, like regex
eos_token = "$"  # end of sentence token, like regex
characters = CHARACTERS + sos_token + eos_token
n_characters = len(characters)
eos_index = characters.find(eos_token)  # end of sequence index
sos_index = characters.find(sos_token)  # start of sequence index

# model hyperparameters
dropout_rate = 0.1
n_rnn_layers = 3  # number of hidden layers
n_hidden = 128  # number of units in hidden layer
criterion = nn.NLLLoss()  # negative log likelihood loss
learning_rate = 0.0005


class CodeGeneratorRNN(nn.Module):

    """Simple RNN module

    Adapted from:
    http://pytorch.org/tutorials/intermediate/char_rnn_generation_tutorial.html
    """

    def __init__(
            self, metafeature_size, input_size, hidden_size, output_size,
            dropout_rate=0.1, num_rnn_layers=1):
        super(CodeGeneratorRNN, self).__init__()
        self.dropout_rate = dropout_rate
        self.num_rnn_layers = num_rnn_layers
        self.hidden_size = hidden_size
        self.output_size = output_size
        self.metafeature_size = metafeature_size
        self.rnn = nn.GRU(
            metafeature_size + input_size, hidden_size,
            num_layers=self.num_rnn_layers)
        self.decoder = nn.Linear(hidden_size, output_size)
        self.dropout = nn.Dropout(self.dropout_rate)
        self.softmax = nn.LogSoftmax(1)

    def forward(self, metafeatures, input, hidden):
        input_concat = torch.cat((metafeatures, input), 2)
        output, hidden = self.rnn(input_concat, hidden)
        output = self.dropout(self.decoder(output))
        output = output.view(output.shape[0], -1)  # dim <seq_length x n_chars>
        output = self.softmax(output)
        return output, hidden

    def initHidden(self):
        return Variable(torch.zeros(self.num_rnn_layers, 1, self.hidden_size))


def create_training_data(algorithm_env, n=5):
    """Samples from algorithm env and creates 4 variants per sample.

    - executable code sample (Estimator class)
    - executable code sample (Estimator instance)
    - non-executable partially randomized code
    - non-executable fully randomized code
    """
    training_data = []
    for i in range(n):
        sample_data = algorithm_env.sample_algorithm_code()
        training_data.extend([
            sample_data,
            algorithm_env.algorithm_obj_to_instance(sample_data),
        ])
    return training_data


def create_metafeature_tensor(metafeatures, seq):
    """Convert a metafeature vector into a tensor.

    For now this will be a single category indicating `is_executable`.

    :returns Tensor: dim <string_length x 1 x total_num_categories>, where
        total_num_categories accounts for all categorical values of the
        metafeatures.
    """
    m = []
    for i, f in enumerate(metafeatures):
        t = torch.zeros(len(seq), 1, len(metafeature_categories[i]))
        cat_index = metafeature_categories[i].index(f)
        for j, _ in enumerate(seq):
            t[j][0][cat_index] = 1
        m.append(t)
    m = torch.cat(m, 2)
    return m


def create_input_tensor(seq):
    """Convert a string of characters to an input tensor.

    Returns tensor of dim <string_length x 1 x n_characters>."""
    t = torch.zeros(len(seq), 1, n_characters)
    for i, c in enumerate(seq):
        t[i][0][characters.find(c)] = 1
    return t


def create_target_tensor(seq):
    """Convert a string of character to a target tensor."""
    char_indices = [characters.find(seq[i]) for i in range(1, len(seq))]
    char_indices.append(eos_index)  # add <EOS> token
    return torch.LongTensor(char_indices)


def _prepend_sos_token(input_seq):
    """Add sos token to the beginning of sequence."""
    return sos_token + input_seq


def random_training_example(metafeatures, input_seq):
    """Sample a random input, target pair."""
    input_seq = _prepend_sos_token(input_seq)
    metafeature_tensor = Variable(
        create_metafeature_tensor(metafeatures, input_seq))
    input_tensor = Variable(create_input_tensor(input_seq))
    target_tensor = Variable(create_target_tensor(input_seq))
    return metafeature_tensor, input_tensor, target_tensor


def train(rnn, optim, metafeature_tensor, input_tensor, target_tensor):
    """Train the character-level model."""
    hidden = rnn.initHidden()
    rnn.zero_grad()
    loss = 0

    output, hidden = rnn(metafeature_tensor, input_tensor, hidden)
    output = output.view(output.shape[0], -1)
    loss += criterion(output, target_tensor)

    loss.backward()
    optim.step()

    return output, loss.data[0] / input_tensor.size()[0]


def time_since(since):
    now = time.time()
    s = now - since
    m = math.floor(s / 60)
    s -= m * 60
    return "%dm %ds" % (m, s)


def sample_rnn(rnn, start_char, metafeatures, max_length=100):
    """Sample from generator given a starting character."""
    metafeature_tensor = Variable(
        create_metafeature_tensor(metafeatures, start_char))
    input_tensor = Variable(create_input_tensor(start_char))
    hidden = rnn.initHidden()

    output_sample = ""

    for i in range(max_length):
        output, hidden = rnn(metafeature_tensor, input_tensor, hidden)
        topv, topi = output.data.topk(1)
        topi = topi[0][0]
        if topi == eos_index:  # <EOS> token
            break
        else:
            char = characters[topi]
            output_sample += char
        input_tensor = Variable(create_input_tensor(char))

    return output_sample


def generate_samples(rnn, start_chars=sos_token * 3, metafeatures=None):
    """"""
    if metafeatures is None:
        metafeatures = [
            ["executable", "not_creates_estimator"]
            for _ in range(len(start_chars))]
    return [sample_rnn(rnn, c, m) for c, m in zip(start_chars, metafeatures)]


def load_model(path):
    rnn = CodeGeneratorRNN(
        n_metafeatures, n_characters, n_hidden, n_characters,
        dropout_rate=dropout_rate, num_rnn_layers=n_rnn_layers)
    rnn.load_state_dict(torch.load(path))
    return rnn



# TODO: these should be enum
ONE_HOT_ENCODER = "one_hot_encoder"
IMPUTER = "imputer"
RESCALER = "rescaler"
FEATURE_PREPROCESSOR = "feature_preprocessor"
CLASSIFIER = "classifier"
REGRESSOR = "regressor"
ALGORITHM_TYPES = [
    ONE_HOT_ENCODER,
    IMPUTER,
    RESCALER,
    FEATURE_PREPROCESSOR,
    CLASSIFIER,
    REGRESSOR,
]
START_TOKEN = "<sos>"
END_TOKEN = "<eos>"
NONE_TOKEN = "<none>"



"""Warnings to ignore when importing package dependencies."""

import warnings

warnings.filterwarnings(
    "ignore", message="numpy.core.umath_tests is an internal NumPy module",
    category=DeprecationWarning)



"""Module to handle logic for data environments."""

import numpy as np
import pandas as pd

from collections import namedtuple, defaultdict
from sklearn.model_selection import train_test_split

from ..data_types import FeatureType, TargetType

from typing import List, Union, Tuple

NULL_DATA_ENV = "<NULL_DATA_ENV>"


PreprocessedFeatures = namedtuple(
    "PreprocessedFeatures", [
        "X",
        "feature_types",
        "feature_indices",
        "feature_names",
    ], defaults=(None, ))


DataEnvSample = namedtuple(
    "DataEnvSample", [
        "X_train",
        "y_train",
        "X_validation",
        "y_validation",
    ])


def create_simple_date_features(x, name=None):
    """Create of simple date features.

    Creates five` numerical features from a datetime:
    - year
    - month-of-year
    - day-of-year
    - day-of-month
    - day-of-week
    """
    x_dates = pd.to_datetime(x)
    x = [
        x_dates.year.values.reshape(-1, 1),
        x_dates.month.values.reshape(-1, 1),
        x_dates.dayofyear.values.reshape(-1, 1),
        x_dates.day.values.reshape(-1, 1),  # month
        x_dates.dayofweek.values.reshape(-1, 1),
    ]
    if name is not None:
        feature_names = [
            f"{name}_year",
            f"{name}_month",
            f"{name}_day_of_year",
            f"{name}_day",
            f"{name}_day_of_week",
        ]
    else:
        feature_names = None
    return x, feature_names


def _stack_features(
        clean_features: List[np.ndarray],
        feature_types: List[FeatureType],
        feature_names: List[str]=None
        ) -> PreprocessedFeatures:
    """Stack continuous features before categorical features."""
    if feature_names is not None and len(feature_names) != len(feature_types):
        raise ValueError(
            "expected feature_types and feature_names to have the same "
            f"length, found {len(feature_names)} and {len(feature_types)}")
    feature_groups = defaultdict(list)
    feature_name_groups = defaultdict(list)
    for i, (ftype, x) in enumerate(zip(feature_types, clean_features)):
        # all feature types should reduce to categorical and continuous
        feature_groups[ftype].append(x)
        if feature_names is not None:
            feature_name_groups[ftype].append(feature_names[i])

    X, output_feature_types = [], []
    output_feature_names = None if feature_names is None else []
    for ftype in [FeatureType.CONTINUOUS, FeatureType.CATEGORICAL]:
        X.extend(feature_groups[ftype])
        output_feature_types.extend([ftype] * len(feature_groups[ftype]))
        if feature_names is not None:
            output_feature_names.extend(feature_name_groups[ftype])

    return PreprocessedFeatures(
        np.hstack(X),
        output_feature_types,
        list(range(len(output_feature_types))),
        output_feature_names,
    )


def preprocess_features(
        features: np.ndarray,
        feature_types: List[FeatureType],
        feature_names: List[str]=None,
        ) -> PreprocessedFeatures:
    """Prepare data environment for use by task environment.

    :param numpy.array features: rows are training instances and columns are
        features.
    :param list[FeatureType] feature_types: specifies feature types per column.
    """
    clean_features = []
    clean_feature_types = []
    clean_feature_names = None if feature_names is None else []

    for i, ftype in enumerate(feature_types):
        x_i = features[:, i]
        if ftype == FeatureType.CATEGORICAL:
            clean_x = [x_i]
            clean_feature_types.append(FeatureType.CATEGORICAL)
            if feature_names is not None:
                clean_feature_names.append(feature_names[i])

        elif ftype == FeatureType.DATE:
            clean_x, date_features = create_simple_date_features(
                x_i, None if feature_names is None else feature_names[i])
            clean_feature_types.extend(
                [FeatureType.CONTINUOUS for _ in range(len(clean_x))])
            if feature_names is not None:
                clean_feature_names.extend(date_features)

        elif ftype == FeatureType.CONTINUOUS:
            clean_x = [x_i.astype(float)]
            clean_feature_types.append(FeatureType.CONTINUOUS)
            if feature_names is not None:
                clean_feature_names.append(feature_names[i])

        if len(clean_x) == 1 and len(clean_x[0].shape) == 1:
            clean_x = [clean_x[0].reshape(-1, 1)]
        clean_features.extend(clean_x)

    assert len(clean_features) == len(clean_feature_types)
    return _stack_features(
        clean_features, clean_feature_types, clean_feature_names)


def _create_data_partition_fns(fetch_training_data, test_size, random_state):
    X_train, X_test, y_train, y_test = train_test_split(
        *fetch_training_data(), test_size=test_size, random_state=random_state)

    # TODO: figure out a way of doing this lazily without having to actually
    # fetch the training data when data environment is initialized

    def _fetch_training_data():
        return X_train, y_train

    def _fetch_test_data():
        return X_test, y_test

    return _fetch_training_data, _fetch_test_data


class DataEnvironment(object):

    def __init__(
            self, name, source, target_type, feature_types, feature_indices,
            fetch_training_data, fetch_test_data=None, test_size=None,
            random_state=None, target_preprocessor=None, scorer=None):
        """Initialize data environment, used to sample tasks.

        :param str name: name of data env.
        :param DataSourceType source: where data environment was obtained.
        :param TargetType target_type: the target type of data env.
        :param list[FeatureType] feature_types: type of each feature.
        :param list[int] feature_indices: indices in the feature design matrix
            corresponding to the feature type.
        :param callable fetch_training_data: function with no arguments that
            returns an array of X, y where X is the feature matrix and y is
            the target matrix.
        :param callable fetch_test_data: function with no arguments that
            returns an array of X, y where X is the feature matrix and y is
            the target matrix.
        :param float test_size: Number of examples to withhold from the
            training set. This option only works when fetch_test_data is None.
        :param int random_state: random state to use when selecting holdout
            test-set. Only applicable when fetch_test_data is None.
        :param transformer target_preprocessor: an sklearn-compliant target
            preprocessor that has a fit/transform/fit_transform method. This
            is applied to the target

        TODO: add `is_sparse` indicator attribute
        """
        self.name = name
        self.source = source
        self.target_type = target_type
        self.raw_feature_types = feature_types
        self.raw_feature_indices = feature_indices
        if test_size and not 0 < test_size < 1:
            raise ValueError(
                "test_size must be a float between 0 and 1, found: %0.02f"
                % test_size)
        self.test_size = test_size
        self.random_state = random_state
        if fetch_test_data is None and self.test_size is not None:
            fetch_training_data, fetch_test_data = _create_data_partition_fns(
                fetch_training_data, self.test_size, self.random_state)
        self.fetch_training_data = fetch_training_data
        self.fetch_test_data = fetch_test_data
        self.target_preprocessor = target_preprocessor
        self._scorer = scorer
        self._data_cache = {}

        # set by cache_fetch_data
        self.feature_types = None
        self.feature_indices = None

    @property
    def training_data(self):
        return self.cache_fetch_data("train")

    @property
    def test_data(self):
        if self.fetch_test_data is None:
            raise RuntimeError(
                "fetch_test_data attribute is None. Cannot produce a "
                "test set. You need to provide a fetch_test_data function "
                "to the DataEnvironment initializer, or specify the test_size "
                "argument.")
        return self.cache_fetch_data("test")

    @property
    def classes(self):
        if self.target_type is TargetType.MULTICLASS:
            return sorted(list(set(self.training_data[1])))
        return None

    @property
    def X_test(self):
        return self.test_data[0]

    @property
    def y_test(self):
        return self.test_data[1]

    @property
    def scorer(self):
        return self._scorer

    def cache_fetch_data(self, partition):
        feature_key = "%s_features" % partition
        target_key = "%s_target" % partition
        if partition == "train":
            fetch_data_fn = self.fetch_training_data
        elif partition == "test":
            fetch_data_fn = self.fetch_test_data
        else:
            raise ValueError(
                "partition should be 'train' or 'test', found: %s" % partition)
        if feature_key not in self._data_cache and \
                target_key not in self._data_cache:
            _features, _target = fetch_data_fn()
            if self.target_preprocessor is not None:
                _target = self.target_preprocessor().fit_transform(_target)
            _features, feature_types, feature_indices, _ = preprocess_features(
                _features, self.raw_feature_types, feature_names=None)

            # cache features and target
            self._data_cache[feature_key] = _features
            self._data_cache[target_key] = _target

            # set feature types and indices based on the training partition
            if partition == "train":
                self.feature_types = feature_types
                self.feature_indices = feature_indices
        return self._data_cache[feature_key], self._data_cache[target_key]

    def sample(self, n=None):
        features, target = self.training_data
        if n and features.shape[0] > n:
            # select subset of training data (mainly for large datasets)
            idx = np.random.choice(range(features.shape[0]), n)
            features, target = features[idx], target[idx]

        data_idx = np.array(range(features.shape[0]))
        # bootstrap sampling
        # TODO: support other kinds of data sampling, e.g. cross validation
        train_index = np.random.choice(data_idx, len(data_idx), replace=True)
        validation_index = np.setdiff1d(data_idx, train_index)
        return DataEnvSample(
            features[train_index],
            target[train_index],
            features[validation_index],
            target[validation_index],
        )



"""Module for handling errors occurring in the TaskEnvironment runtime.

This module handles the validity of exceptions raised by the TaskEnvironment
when fitting, predicting, and scoring a propoosed ML framework.
"""

import re

from scipy.optimize.optimize import LineSearchWarning
from scipy.linalg import LinAlgWarning
from sklearn.exceptions import ConvergenceWarning, UndefinedMetricWarning


# sklearn error messages that yield negative reward
# TODO: need to add unit tests for this module
# TODO: characterize the root of each of these error messages.
FIT_ERROR_MESSAGES = [
    (AttributeError,
        "module 'scipy.cluster._hierarchy' has no attribute 'nn_chain'"),
    (ValueError, "Cannot extract more clusters than samples"),
    (ValueError, "l2 was provided as affinity."),
    (ValueError,
        "l1 was provided as affinity. Ward can only work with euclidean "
        "distances"),
    (ValueError,
        "manhattan was provided as affinity. Ward can only work with "
        "euclidean distances"),
    (ValueError,
        "cosine was provided as affinity. Ward can only work with "
        "euclidean distances"),
    (ValueError, "No feature in X meets the variance threshold"),
    (ValueError,
        "The condensed distance matrix must contain only finite values"),
    (ValueError, "n_components must be < n_features"),
    (ValueError, r"max_features must be in \(0, n_features\]"),
    (ValueError,
        r"Input contains NaN, infinity or a value too large for "
        r"dtype\('float64'\)"),
    (ValueError, "Input X must be non-negative"),
    (FloatingPointError, "^overflow encountered in"),
    (FloatingPointError, "^underflow encountered in"),
    (FloatingPointError, "^divide by zero encountered in"),
    (FloatingPointError, "^invalid value encountered in"),
    (FloatingPointError, "^underflow encountered in"),
    (TypeError, "'str' object is not callable"),
    (TypeError, "'numpy.float64' object cannot be interpreted as an integer"),
    (ZeroDivisionError,
        "Current sag implementation does not handle the case step_size"),
]
FIT_ERROR_TYPES = tuple(set([i for i, _ in FIT_ERROR_MESSAGES]))


# these should be ignored during fitting time so that the mlf can still
# generate validation scores for these
FIT_WARNINGS = [
    (ConvergenceWarning,
        "The max_iter was reached which means the coef_ did not converge"),
    (ConvergenceWarning,
        "newton-cg failed to converge. Increase the number of iterations"),
    (ConvergenceWarning,
        "The max_iter was reached which means the coef_ did not converge"),
    (ConvergenceWarning,
        "Objective did not converge. You might want to increase the number of "
        "iterations. Fitting data with very small alpha may cause precision "
        "problems."),
    (ImportWarning,
        "can't resolve package from __spec__ or __package__, falling back on "
        "__name__ and __path__"),
    (LineSearchWarning, "The line search algorithm did not converge"),
    (LinAlgWarning,
        "scipy.linalg.solve\nIll-conditioned matrix detected. Result "
        "is not guaranteed to be accurate.\nReciprocal condition "
        "number"),
    (UserWarning, "Line Search failed"),
    (UserWarning, "n_components > n_samples. This is not possible."),
    (UserWarning, "n_components is too large: it will be set to"),
    (UserWarning, "Ignoring n_components with whiten=False."),
    (UserWarning,
        "FastICA did not converge. Consider increasing tolerance or "
        "the maximum number of iterations."),
    (UserWarning,
        "Singular matrix in solving dual problem. Using least-squares "
        "solution instead."),
    (RuntimeWarning,
        "numpy.dtype size changed, may indicate binary incompatibility"),
]


PREDICT_ERROR_MESSAGES = [
    (FloatingPointError, "overflow encountered in exp"),
    (FloatingPointError, "overflow encountered in multiply"),
    (FloatingPointError, "overflow encountered in power"),
    (FloatingPointError, "overflow encountered in reduce"),
    (FloatingPointError, "divide by zero encountered in log"),
    (FloatingPointError, "divide by zero encountered in true_divide"),
    (FloatingPointError, "divide by zero encountered in double_scalars"),
    (FloatingPointError, "underflow encountered in true_divide"),
    (FloatingPointError, "underflow encountered in power"),
    (FloatingPointError, "underflow encountered in reciprocal"),
    (FloatingPointError, "underflow encountered in exp"),
    (FloatingPointError, "underflow encountered in multiply"),
    # SimpleImputer algorithm component
    (ValueError,
        r"The features \[.+\] have missing values in transform "
        "but have no missing values in fit"),
    # KNearestNeighbors algorithm component
    (ValueError,
        r"^Expected n_neighbors <= n_samples,")
]

PREDICT_ERROR_TYPES = tuple(set([i for i, _ in PREDICT_ERROR_MESSAGES]))

SCORE_ERRORS = (ValueError, )
SCORE_WARNINGS = [
    (UndefinedMetricWarning,
        "F-score is ill-defined and being set to 0.0 in labels with no "
        "predicted samples"),
    (UndefinedMetricWarning,
        "F-score is ill-defined and being set to 0.0 in labels with no true "
        "samples."),
    (RuntimeWarning, "invalid value encountered in log"),
]


def _is_valid_error(error, error_message_tuples):
    error_str = str(error)
    for etype, msg in error_message_tuples:
        if isinstance(error, etype) and re.match(msg, error_str):
            return True
    return False


def is_valid_fit_error(error):
    """Return True if error in MLF fit is valid for controller training.

    The criterion for acceptability is somewhat tautological at this point,
    essentially defined as errors that result from things like poorly specified
    hyperparameters and other errors raised by the sklearn API.

    Things that are not explicitly listed in the FIT_ERROR_MESSAGES global
    variable are considered invalid fit errors. This helps to minimize
    unexpected behavior when training the CASH controller.

    :params BaseException error: any subclass of BaseException.
    :returns: True if valid fit error, False otherwise.
    """
    return isinstance(error, FIT_ERROR_TYPES) and \
        _is_valid_error(error, FIT_ERROR_MESSAGES)


def is_valid_predict_error(error):
    """Return True if error in MLF predict is valid for controller training."""
    return isinstance(error, PREDICT_ERROR_TYPES) and \
        _is_valid_error(error, PREDICT_ERROR_MESSAGES)



"""Regressor components."""

from sklearn.ensemble import AdaBoostRegressor, RandomForestRegressor
from sklearn.gaussian_process import GaussianProcessRegressor
from sklearn.linear_model import ARDRegression, BayesianRidge, Lasso, Ridge
from sklearn.neighbors import KNeighborsRegressor
from sklearn.svm import LinearSVR, SVR
from sklearn.tree import DecisionTreeRegressor

from .algorithm_component import AlgorithmComponent, EXCLUDE_ALL
from .hyperparameter import (
    CategoricalHyperparameter, UniformIntHyperparameter,
    UniformFloatHyperparameter, BaseEstimatorHyperparameter)
from . import constants


# TODO: add sgd regressor, xgboost regressor


# ===================
# Ensemble Regressors
# ===================

def adaboost_regression():
    return AlgorithmComponent(
        name="AdaBoostRegressor",
        component_class=AdaBoostRegressor,
        component_type=constants.REGRESSOR,
        hyperparameters=[
            BaseEstimatorHyperparameter(
                hname="base_estimator",
                base_estimator=DecisionTreeRegressor,
                hyperparameters=[
                    UniformIntHyperparameter("max_depth", 1, 10, default=1)
                ],
                default=DecisionTreeRegressor(max_depth=1)
            ),
            UniformIntHyperparameter(
                "n_estimators", 50, 500, default=50, n=10),
            UniformFloatHyperparameter(
                "learning_rate", 0.1, 2., default=1., log=True, n=10),
            CategoricalHyperparameter(
                "loss", ["linear", "square", "exponential"], default="linear")
        ])


def extra_trees_regression():
    pass


def gradient_boosting_regression():
    pass


def random_forest_regression():
    return AlgorithmComponent(
        name="RandomForestRegressor",
        component_class=RandomForestRegressor,
        component_type=constants.REGRESSOR,
        hyperparameters=[
            UniformIntHyperparameter(
                "n_estimators", 50, 200, default=50, n=10),
            CategoricalHyperparameter(
                "criterion", ["mse", "friedman_mse", "mae"], default="mse"),
            UniformFloatHyperparameter(
                "max_features", 0.1, 1.0, default=1.0),
            UniformIntHyperparameter(
                "min_samples_split", 2, 20, default=2),
            UniformIntHyperparameter(
                "min_samples_leaf", 1, 20, default=1),
            CategoricalHyperparameter(
                "bootstrap", [True, False], default=True),
        ],
        constant_hyperparameters={
            "max_depth": None,
            "min_weight_fraction_leaf": 0.0,
            "max_leaf_nodes": None,
            "min_impurity_decrease": 0.0,
        })


# ===================
# Gaussian Regressors
# ===================

def rbf_gaussian_process_regression():
    """Create gaussian process regressorw with an RBF kernel.

    Use the default "1.0 * RBF(1.0)" kernel.

    Auto-sklearn implements the kernel as dependent on `n_features` of the
    training data:
    https://github.com/automl/auto-sklearn/blob/master/autosklearn/pipeline/components/regression/gaussian_process.py  # noqa

    Later need to decide whether it's worth supporting this kind of
    implementation by dynamically instantiating the estimator based on the
    data environment.
    """
    return AlgorithmComponent(
        name="GaussianProcessRegressor",
        component_class=GaussianProcessRegressor,
        component_type=constants.REGRESSOR,
        hyperparameters=[
            UniformFloatHyperparameter(
                "alpha", 1e-14, 1.0, default=1e-8, log=True),
        ],
        constant_hyperparameters={
            "optimizer": "fmin_l_bfgs_b",
            "n_restarts_optimizer": 10,
            "normalize_y": True,
        })


# =================
# Linear Regressors
# =================

def ard_regression():
    return AlgorithmComponent(
        name="ARDRegression",
        component_class=ARDRegression,
        component_type=constants.REGRESSOR,
        hyperparameters=[
            UniformFloatHyperparameter(
                "tol", 1e-5, 1e-1, default=1e-3, log=True),
            UniformFloatHyperparameter(
                "alpha_1", 1e-10, 1e-3, default=1e-6, log=True),
            UniformFloatHyperparameter(
                "alpha_2", 1e-10, 1e-3, default=1e-6, log=True),
            UniformFloatHyperparameter(
                "lambda_1", 1e-10, 1e-3, default=1e-6, log=True),
            UniformFloatHyperparameter(
                "lambda_2", 1e-10, 1e-3, default=1e-6, log=True),
            UniformFloatHyperparameter(
                "threshold_lambda", 10 ** 3, 10 ** 5, default=10 ** 4,
                log=True),
        ],
        constant_hyperparameters={
            "n_iter": 300,
            "fit_intercept": True,
            "compute_score": False
        })


def bayesian_ridge_regression():
    return AlgorithmComponent(
        name="BayesianRidge",
        component_class=BayesianRidge,
        component_type=constants.REGRESSOR,
        hyperparameters=[
            UniformFloatHyperparameter(
                "tol", 1e-5, 1e-1, default=1e-3, log=True),
            UniformFloatHyperparameter(
                "alpha_1", 1e-10, 1e-3, default=1e-6, log=True),
            UniformFloatHyperparameter(
                "alpha_2", 1e-10, 1e-3, default=1e-6, log=True),
            UniformFloatHyperparameter(
                "lambda_1", 1e-10, 1e-3, default=1e-6, log=True),
            UniformFloatHyperparameter(
                "lambda_2", 1e-10, 1e-3, default=1e-6, log=True),
        ],
        constant_hyperparameters={
            "n_iter": 300,
            "fit_intercept": True,
            "compute_score": False
        })


def ridge_regression():
    """Create linear ridge regression algorithm component."""
    return AlgorithmComponent(
        name="RidgeRegression",
        component_class=Ridge,
        component_type=constants.REGRESSOR,
        hyperparameters=[
            UniformFloatHyperparameter(
                "alpha", 10 ** -5, 10.0, default=1., log=True),
            UniformFloatHyperparameter(
                "tol", 1e-5, 1e-1, default=1e-3, log=True)
        ],
        constant_hyperparameters={
            "fit_intercept": True,
        })


def lasso_regression():
    """Create linear lasso regression algorithm component."""
    return AlgorithmComponent(
        name="LassoRegression",
        component_class=Lasso,
        component_type=constants.REGRESSOR,
        hyperparameters=[
            UniformFloatHyperparameter(
                "alpha", 10 ** -5, 10.0, default=1., log=True),
            UniformFloatHyperparameter(
                "tol", 1e-5, 1e-1, default=1e-3, log=True)
        ],
        constant_hyperparameters={
            "fit_intercept": True,
        })


def elastic_net_regression():
    pass


def sgd_regression():
    pass


# =========================
# Non-parametric Regressors
# =========================

def k_nearest_neighbors_regression():
    return AlgorithmComponent(
        name="KNearestNeighorsRegressor",
        component_class=KNeighborsRegressor,
        component_type=constants.REGRESSOR,
        hyperparameters=[
            UniformIntHyperparameter(
                "n_neighbors", 1, 100, log=True, default=1),
            CategoricalHyperparameter(
                "weights", ["uniform", "distance"], default="uniform"),
            CategoricalHyperparameter("p", [1, 2], default=2)
        ])


# =========================
# Support-vector Regressors
# =========================


def support_vector_regression_linear():
    # just following the parameterization used by auto-sklearn:
    # https://github.com/automl/auto-sklearn/blob/master/autosklearn/pipeline/components/regression/liblinear_svr.py  # noqa
    return AlgorithmComponent(
        name="LinearSVR",
        component_class=LinearSVR,
        component_type=constants.REGRESSOR,
        hyperparameters=[
            UniformFloatHyperparameter(
                "C", 0.03125, 32768, default=1.0, log=True, n=10),
            UniformFloatHyperparameter(
                "epsilon", 0.001, 1, default=0.1, log=True, n=10),
            UniformFloatHyperparameter(
                "tol", 1e-5, 1e-1, default=1e-4, log=True, n=10),
        ],
        constant_hyperparameters={
            "loss": "squared_epsilon_insensitive",
            "dual": False,
            "fit_intercept": True,
            "intercept_scaling": 1,
        })


def support_vector_regression_nonlinear():
    return AlgorithmComponent(
        name="NonlinearSVR",
        component_class=SVR,
        component_type=constants.REGRESSOR,
        hyperparameters=[
            UniformFloatHyperparameter(
                "C", 0.03125, 32768, default=1.0, log=True),
            UniformFloatHyperparameter(
                "epsilon", 0.001, 1, default=0.1, log=True),
            CategoricalHyperparameter(
                "kernel", ["rbf", "poly", "sigmoid"], default="rbf"),
            UniformIntHyperparameter("degree", 2, 5, default=3),
            UniformFloatHyperparameter(
                "gamma", 3.0517578125e-05, 8, default=0.1, log=True),
            UniformFloatHyperparameter("coef0", -1, -1, default=0),
            CategoricalHyperparameter(
                "shrinking", [True, False], default=True),
            UniformFloatHyperparameter(
                "tol", 1e-5, 1e-1, default=1e-3, log=True),
        ],
        constant_hyperparameters={
            "max_iter": -1,
            "cache_size": 200,
        },
        exclusion_conditions={
            "kernel": {
                "rbf": {
                    "coef0": EXCLUDE_ALL,
                    "degree": EXCLUDE_ALL,
                },
                "sigmoid": {
                    "degree": EXCLUDE_ALL
                }
            }
        })


def _libsvm_hyperparameters():
    return [
        UniformFloatHyperparameter(
            "C", 0.03125, 32768, default=1.0, log=True),
        UniformFloatHyperparameter(
            "epsilon", 0.001, 1, default=0.1, log=True),
        UniformFloatHyperparameter(
            "gamma", 3.0517578125e-05, 8, default=0.1, log=True),
        CategoricalHyperparameter(
            "shrinking", [True, False], default=True),
        UniformFloatHyperparameter(
            "tol", 1e-5, 1e-1, default=1e-3, log=True),
    ]


def _libsvm_poly_sigmoid_hyperparameters():
    return [
        UniformFloatHyperparameter("coef0", -1, -1, default=0)
    ]


def _libsvm_constant_hyperparameters():
    return {
        "max_iter": -1,
        "cache_size": 200,
    }


def support_vector_regression_poly():
    constant_hyperparameters = {"kernel": "poly"}
    constant_hyperparameters.update(_libsvm_constant_hyperparameters())
    return AlgorithmComponent(
        name="PolyKernelSVR",
        component_class=SVR,
        component_type=constants.REGRESSOR,
        hyperparameters=(
            _libsvm_hyperparameters() +
            _libsvm_poly_sigmoid_hyperparameters() +
            [UniformIntHyperparameter("degree", 2, 5, default=3)]
        ),
        constant_hyperparameters=constant_hyperparameters,
    )


def support_vector_regression_rbf():
    constant_hyperparameters = {"kernel": "rbf"}
    constant_hyperparameters.update(_libsvm_constant_hyperparameters())
    return AlgorithmComponent(
        name="RBFKernelSVR",
        component_class=SVR,
        component_type=constants.REGRESSOR,
        hyperparameters=_libsvm_hyperparameters(),
        constant_hyperparameters=constant_hyperparameters
    )


def support_vector_regression_sigmoid():
    constant_hyperparameters = {"kernel": "sigmoid"}
    constant_hyperparameters.update(_libsvm_constant_hyperparameters())
    return AlgorithmComponent(
        name="SigmoidKernelSVR",
        component_class=SVR,
        component_type=constants.REGRESSOR,
        hyperparameters=(
            _libsvm_hyperparameters() +
            _libsvm_poly_sigmoid_hyperparameters()
        ),
        constant_hyperparameters=constant_hyperparameters,
    )


# =====================
# Tree-based Regressors
# =====================

def decision_tree_regression():
    """Create decision tree regressor algorithm component."""
    return AlgorithmComponent(
        name="DecisionTreeRegressor",
        component_class=DecisionTreeRegressor,
        component_type=constants.REGRESSOR,
        hyperparameters=[
            CategoricalHyperparameter(
                "criterion", ["mse", "friedman_mse", "mae"], default="mse"),
            CategoricalHyperparameter(
                "splitter", ["best", "random"], default="best"),
            UniformIntHyperparameter("max_depth", 1, 4, default=1, n=4),
            UniformIntHyperparameter(
                "min_samples_split", 2, 20, default=2, n=10),
            UniformIntHyperparameter(
                "min_samples_leaf", 1, 20, default=1, n=10),
        ],
        constant_hyperparameters={
            "min_weight_fraction_leaf": 0.0,
            "max_features": 1,
            "max_leaf_nodes": None,
            "min_impurity_decrease": 0.0,
        })



"""Define scoring functions."""

import numpy as np
import operator
import sklearn.metrics

from collections import namedtuple
from functools import partial

# TODO: full coverage of sklearn.metrics


Scorer = namedtuple(
    "Scorer", [
        "name",
        "fn",  # sklearn-compliant scoring function e.g. mean_squared_error
        "reward_transformer",  # function that bounds range of scoring function
        # function to determine if score a is better than b. Has signature
        # x, y -> bool, returning true if x is better than y.
        "comparator",
        "needs_proba",
    ])

RegressionScorer = partial(Scorer, needs_proba=False)


def multiclass_classification_metrics():
    return [
        accuracy(),
        precision(),
        recall(),
        f1_score_weighted_average(),
        f1_score_macro(),
        log_loss(),
    ]


def binary_classification_metrics():
    return multiclass_classification_metrics() + [roc_auc()]


def regression_metrics():
    return [
        mean_absolute_error(),
        mean_squared_error(),
        r2_score(),
        root_mean_squared_error(),
        root_mean_squared_log_error(),
    ]


def exponentiated_log(x, gamma=0.1):
    """Bounds functions with a range of >= 0 to range of [0, 1].

    This function is used for scorers where larger scores mean worse
    performance. This function returns 1.0 if the score is 0, and returns
    values that asymptotically approach 0 as the score approaches positive
    infinity.

    :param float x: value to transform
    :param float gamma: decay coefficient. Larger gamma means function decays
        more quickly as x gets larger.
    :returns: transformed value.
    :rtype: float
    """
    if x < 0:
        raise ValueError("value %s not a valid input. Must be >= 0")
    if x == 0:
        # since the below function is undefined at x=0, return 1 if x=0.
        return 1.0
    return 1 / (1 + np.power(np.e, np.log(gamma * x)))


def rectified_linear(x):
    """Bounds functions with a range of [-infinity, 1] to [0, 1].

    Negative values are clipped to 0.

    This function is used for scorers, e.g. R^2 (regression metric), where
    1 is the best possible score, 0 indicates expected value of y, and negative
    numbers are worse than predicting the mean.
    """
    return 0 if x < 0 else x


# CLASSIFICATION METRICS

def accuracy():
    """Accuracy scorer."""
    return Scorer(
        name="accuracy",
        fn=sklearn.metrics.accuracy_score,
        reward_transformer=None,
        comparator=operator.gt,
        needs_proba=False)


def precision():
    return Scorer(
        name="precision",
        fn=sklearn.metrics.precision_score,
        reward_transformer=None,
        comparator=operator.gt,
        needs_proba=False)


def recall():
    return Scorer(
        name="recall",
        fn=sklearn.metrics.recall_score,
        reward_transformer=None,
        comparator=operator.gt,
        needs_proba=False)


def f1_score_weighted_average():
    """F1 score weighted average scorer."""
    return Scorer(
        name="f1_weighted_average",
        fn=partial(sklearn.metrics.f1_score, average="weighted"),
        reward_transformer=None,
        comparator=operator.gt,
        needs_proba=False)


def f1_score_macro():
    """F1 score macro scorer."""
    return Scorer(
        name="f1_macro",
        fn=partial(sklearn.metrics.f1_score, average="macro"),
        reward_transformer=None,
        comparator=operator.gt,
        needs_proba=False)


def log_loss():
    """Logistic loss (cross-entropy loss) scorer."""
    return Scorer(
        name="log_loss",
        fn=sklearn.metrics.log_loss,
        reward_transformer=exponentiated_log,
        comparator=operator.lt,
        needs_proba=True)


def roc_auc():
    """ROC AUC scorer."""
    return Scorer(
        name="roc_auc",
        fn=sklearn.metrics.roc_auc_score,
        reward_transformer=None,
        comparator=operator.gt,
        needs_proba=True)


# REGRESSION_METRICS

def mean_absolute_error():
    """Mean absolute error scorer."""
    return RegressionScorer(
        name="mean_absolute_error",
        fn=sklearn.metrics.mean_absolute_error,
        reward_transformer=exponentiated_log,
        comparator=operator.lt)


def mean_squared_error():
    """Mean squared error scorer."""
    return RegressionScorer(
        name="mean_squared_error",
        fn=sklearn.metrics.mean_squared_error,
        reward_transformer=exponentiated_log,
        comparator=operator.lt)


def r2_score():
    """R^2 coefficient of determination scorer."""
    return RegressionScorer(
        name="r2",
        fn=sklearn.metrics.r2_score,
        reward_transformer=rectified_linear,
        comparator=operator.gt)


def root_mean_squared_error():
    """Root mean squared error scorer."""
    def _rmse_scorer(*args, **kwargs):
        return np.sqrt(sklearn.metrics.mean_squared_error(*args, **kwargs))

    return RegressionScorer(
        name="root_mean_squared_error",
        fn=_rmse_scorer,
        reward_transformer=exponentiated_log,
        comparator=operator.lt)


def root_mean_squared_log_error():
    """Root mean squared log error."""
    def _rmsle_scorer(*args, **kwargs):
        return np.sqrt(sklearn.metrics.mean_squared_log_error(*args, **kwargs))

    return RegressionScorer(
        name="root_mean_squared_log_error",
        fn=_rmsle_scorer,
        reward_transformer=exponentiated_log,
        comparator=operator.lt)



"""Module for generating algorithm and hyperparameter sequences.

This is a controller that jointly generates machine learning framework
pipelines and their associated hyperparameter settings. It does so via an RNN
that has a separate softmax classifier at each time step, where the classifier
has to predict either an algorithm to add to the pipeline or a hyperparameter
for the most recently chosen algorithm.
"""


import dill
import torch
import torch.nn as nn

from collections import defaultdict, OrderedDict
from pathlib import Path
from torch.autograd import Variable
from torch.distributions import Categorical

from .data_types import CASHComponent
from .components.algorithm_component import EXCLUDE_ALL


class MetaLearnController(nn.Module):
    """RNN module to generate joint algorithm and hyperparameter controller."""

    def __init__(
            self,
            metafeature_size,
            input_size,
            hidden_size,
            output_size,
            a_space,
            metafeature_encoding_size=20,
            aux_reward_size=1,
            dropout_rate=0.1,
            num_rnn_layers=1):
        """Initialize Cash Controller.

        :param int metafeature_size: dimensionality of metafeatures
        :param int input_size: dimensionality of input features (actions)
        :param int hidden_size: dimensionality of hidden RNN layer.
        :param int output_size: dimensionality of output space.
        :param AlgorithmSpace a_space:
        :param int metafeature_encoding_size: dim of metafeatures embedding
        :param float dropout_rate:
        :param int num_rnn_layers:
        :ivar int aux_size: dimensionality of auxiliary inputs (reward and
            action from previous time-step).
        :ivar list[dict] action_classifiers:
        :ivar list[float] log_prob_buffer:
        :ivar list[float] reward buffer:
        """
        super(MetaLearnController, self).__init__()
        self.a_space = a_space
        self.dropout_rate = dropout_rate
        self.num_rnn_layers = num_rnn_layers
        self.metafeature_size = metafeature_size
        self.input_size = input_size

        # action dim for previous reward. Note that to implement meta-learning,
        # the previous action is encoded directly in the input.
        self.aux_reward_size = aux_reward_size
        self.hidden_size = hidden_size
        self.output_size = output_size
        self.metafeature_encoding_size = metafeature_encoding_size

        # an array storing classifiers for algorithm components and
        # hyperparameter settings
        self.action_classifiers = []
        # maps algorithm types to index pointer referring to action classifier
        # for algorithm components of that type.
        self.atype_map = {}
        # maps algorithm component to list of index pointers referring to
        # action classifier for hyperparameters for that component
        self.acomponent_to_hyperparam = defaultdict(list)
        # maps hyperparameter actions to conditional masks that determines what
        # subsequent actions the agent can make.
        self.hyperparam_exclude_conditions = defaultdict(dict)

        # buffers needed for backprop
        self.log_prob_buffer = []
        self.reward_buffer = []
        self.entropy_buffer = []

        # architecture specification
        self.rnn = nn.GRU(
            self.input_size + self.aux_reward_size +
            self.metafeature_encoding_size,
            self.hidden_size,
            num_layers=self.num_rnn_layers,
            dropout=self.dropout_rate)
        self.decoder = nn.Linear(self.hidden_size, self.output_size)
        self.dropout = nn.Dropout(self.dropout_rate)
        # special encoding layer for initial metafeature input
        self.metafeature_dense = nn.Linear(
            self.metafeature_size, self.metafeature_encoding_size)

        # for each algorithm component and hyperparameter value, create a
        # softmax classifier over the number of unique components/hyperparam
        # values.
        all_components = self.a_space.component_dict_from_signature(
            self.a_space.ALL_COMPONENTS)
        idx = 0
        # TODO: this logic should be in algorithm_space
        for atype, acomponents in all_components.items():
            # action classifier
            self._add_action_classifier(
                idx, CASHComponent.ALGORITHM, atype, acomponents, None)
            self.atype_map[atype] = idx
            idx += 1
            for acomponent in acomponents:
                h_state_space = self.a_space.h_state_space([acomponent])
                h_exclude_cond_map = self.a_space.h_exclusion_conditions(
                    [acomponent])
                for hname, hchoices in h_state_space.items():
                    exclude_masks = self._exclusion_condition_to_mask(
                        h_state_space, h_exclude_cond_map.get(hname, None))
                    self._add_action_classifier(
                        idx, CASHComponent.HYPERPARAMETER, hname, hchoices,
                        exclude_masks)
                    self.acomponent_to_hyperparam[acomponent].append(idx)
                    idx += 1
        self.n_actions = len(self.action_classifiers)

    def _exclusion_condition_to_mask(self, h_state_space, exclude):
        # TODO: this can probably be moved to algorithm_space.py
        if exclude is None:
            return None

        exclude_masks = defaultdict(dict)
        for choice, exclude_conditions in exclude.items():
            for hname, exclude_values in exclude_conditions.items():
                if exclude_values == EXCLUDE_ALL:
                    mask = [1 for _ in h_state_space[hname]]
                else:
                    mask = [i in exclude_values for i in h_state_space[hname]]
                exclude_masks[choice].update({hname: torch.ByteTensor(mask)})

        return exclude_masks

    def _add_action_classifier(
            self, action_index, action_type, name, choices,
            exclude_masks):
        """Add action classifier to the nn.Module."""
        dense_attr = "action_dense_%s" % action_index
        softmax_attr = "action_softmax_%s" % action_index
        embedding_attr = "action_embedding_%s" % action_index
        # set layer attributes for action classifiers
        n_choices = len(choices)
        setattr(self, dense_attr, nn.Linear(self.output_size, n_choices))
        setattr(self, softmax_attr, nn.Softmax(1))
        setattr(self, embedding_attr, nn.Embedding(n_choices, self.input_size))
        # keep track of metadata for lookup purposes
        self.action_classifiers.append({
            "action_type": action_type,
            "name": name,
            "choices": choices,
            "dense_attr": dense_attr,
            "softmax_attr": softmax_attr,
            "embedding_attr": embedding_attr,
            "exclude_masks": exclude_masks,
        })

    def forward(self, input, aux, metafeatures, hidden, action_classifier):
        """Forward pass through controller."""
        input_concat = torch.cat([
            input, aux, self.metafeature_dense(metafeatures)], 2)
        rnn_output, hidden = self.rnn(input_concat, hidden)
        rnn_output = self.dropout(self.decoder(rnn_output))

        dense = getattr(self, action_classifier["dense_attr"])
        softmax = getattr(self, action_classifier["softmax_attr"])

        # obtain action probability distribution
        rnn_output = rnn_output.view(rnn_output.shape[0], -1)
        action_probs = softmax(dense(rnn_output))
        return action_probs, hidden

    def decode(self, init_input_tensor, target_type, aux, metafeatures,
               init_hidden):
        """Decode a metafeature tensor to sequence of actions.

        Where the actions are a sequence of algorithm components and
        hyperparameter settings.

        TODO: add unit tests for this method and related methods.
        """
        input_tensor, hidden = init_input_tensor, init_hidden
        actions = []
        # create attribute dictionary that maps hyperparameter names to
        # conditional masks
        self._exclude_masks = {}
        # for each algorithm component type, select an algorithm
        for atype in self.a_space.component_dict_from_target_type(target_type):
            action, input_tensor, hidden = self._decode_action(
                input_tensor, aux, metafeatures, hidden, self.atype_map[atype])
            actions.append(action)
            # each algorithm is associated with a set of hyperparameters
            for h_index in self.acomponent_to_hyperparam[action["choice"]]:
                action, input_tensor, hidden = self._decode_action(
                    input_tensor, aux, metafeatures, hidden, h_index)
                if action is None:
                    continue
                actions.append(action)
        return actions, input_tensor

    def _get_exclusion_mask(self, action_name):

        if action_name in self._exclude_masks:
            mask = ~self._exclude_masks[action_name].view(1, -1)
            return Variable(mask.type(torch.FloatTensor))
        else:
            return None

    def _accumulate_exclusion_mask(self, masks):
        for action_name, m in masks.items():
            mask = self._exclude_masks.get(action_name, None)
            if mask is None:
                self._exclude_masks[action_name] = m
            else:
                self._exclude_masks[action_name] = (m + mask) > 0

    def _decode_action(
            self, input_tensor, aux, metafeatures, hidden, action_index):
        action_classifier = self.action_classifiers[action_index]
        mask = self._get_exclusion_mask(action_classifier["name"])
        if mask is not None and (mask == 0).all():
            # If all actions are masked, then don't select any choice
            return None, input_tensor, hidden
        action_probs, hidden = self.forward(
            input_tensor, aux, metafeatures, hidden, action_classifier)
        action = self._select_action(action_probs, action_classifier, mask)
        input_tensor = self.encode_embedding(
            action_index, action["choice_index"])
        return action, input_tensor, hidden

    def _select_action(self, action_probs, action_classifier, mask):
        """Select action based on action probability distribution."""
        # compute unmasked log probabilitises.
        log_action_probs = action_probs.log()
        if mask is not None:
            action_probs = action_probs * mask
        if (action_probs.data > 0).sum() == 0:
            raise RuntimeError(
                "at least one action probability must be > 0")

        choice_index = Categorical(action_probs).sample()
        _choice_index = int(choice_index.data)
        choice = action_classifier["choices"][_choice_index]

        # for current choice, accumulate exclusion masks if any
        if action_classifier["exclude_masks"] is not None and \
                choice in action_classifier["exclude_masks"]:
            for _, masks in action_classifier["exclude_masks"].items():
                self._accumulate_exclusion_mask(masks)

        return {
            "action_type": action_classifier["action_type"],
            "action_name": action_classifier["name"],
            "choices": action_classifier["choices"],
            "choice_index": _choice_index,
            "choice": choice,
            "log_prob": log_action_probs.index_select(1, choice_index),
            "entropy": -(log_action_probs * action_probs).sum(1, keepdim=True)
        }

    def encode_embedding(self, action_index, choice_index):
        """Encode action choice into embedding at input for next action."""
        action_classifier = self.action_classifiers[action_index]
        embedding = getattr(self, action_classifier["embedding_attr"])
        action_embedding = embedding(
            Variable(torch.LongTensor([choice_index])))
        return action_embedding.view(
            1, action_embedding.shape[0], action_embedding.shape[1])

    def init_hidden(self):
        """Initialize hidden layer with zeros."""
        return Variable(torch.zeros(self.num_rnn_layers, 1, self.hidden_size))

    def init_action(self):
        """Initialize action input state."""
        return Variable(torch.zeros(1, 1, self.input_size))

    @property
    def config(self):
        return {
            "metafeature_size": self.metafeature_size,
            "input_size": self.input_size,
            "hidden_size": self.hidden_size,
            "output_size": self.output_size,
            "a_space": self.a_space,
            "dropout_rate": self.dropout_rate,
            "num_rnn_layers": self.num_rnn_layers,
        }

    def save(self, path):
        """Save weights and configuration."""
        torch.save(
            OrderedDict({
                "config": self.config,
                "weights": self.state_dict(),
            }),
            path,
            pickle_module=dill)

    @classmethod
    def load(cls, path):
        """Load saved controller."""
        model_config = torch.load(path, pickle_module=dill)
        rnn = cls(**model_config["config"])
        rnn.load_state_dict(model_config["weights"])
        return rnn



from . import (
    algorithm_env, algorithm_space, beam_search, metalearn_controller,
    metalearn_reinforce, components, data_environments, data_types, errors,
    experiment, inference, loggers, random_search, rnn_code_generator, scorers,
    task_environment, tracking)


__all__ = [
    algorithm_env,
    algorithm_space,
    beam_search,
    metalearn_controller,
    metalearn_reinforce,
    components,
    data_environments,
    data_types,
    errors,
    inference,
    experiment,
    loggers,
    random_search,
    rnn_code_generator,
    scorers,
    task_environment,
    tracking,
]
